/*
  File name: IGameCtrl.cpp
  Copyright: (C) 2004 - 2008 Tomasz Kazmierczak

  Creation date: 08.05.2004
  Last modification date: 20.06.2008

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

  gameplay control interface
*/

#include <cstdlib>
#include <cassert>
#include <math.h>
#include "IGameCtrl.h"
#include "LoopCtrl.h"

#define MACHINEGUN_FIRE_DELAY 450
#define PLAYER_BURN_DELAY 250
IGameCtrl *createGameCtrlInterface()
{
	return new IGameCtrl;
}

bool IGameCtrl::init(UpdateData *data, GOInterface *objects, COGLClass *ogl, CInfoText *texts)
{
	mObjects = objects;
	mOGL = ogl;
	mSDL = CSDLClass::instance();
	mTexts = texts;
	mConfig = CCytadelaConfig::instance();
	mLocInterface = ILocalization::instance();
	mCameraPosition.x = *(data->pX);
	mCameraPosition.z = *(data->pZ);
	mCameraDirection.x = data->fDirX;
	mCameraDirection.z = data->fDirZ;
	mCameraAngles.y = data->fRotY;
	mMiliseconds = data->uiMiliseconds;
	mPrevPosition[0] = mPosition[0] = data->pPosition[0] = (uint8_t)(*(data->pX) / 100);
	mPrevPosition[1] = mPosition[1] = data->pPosition[1] = (uint8_t)-(*(data->pZ) / 100);
	mCamRay = 35.0f;
	mEnemyRay = 35.0f;
	mAuxVTime = 0;
	mRocketLoadTime = 0;
	mMGFireDiff = MACHINEGUN_FIRE_DELAY; //no firing delay for mashinegun at the begining
	mPlayerBurnDiff = PLAYER_BURN_DELAY; //the same for player burn
	mLastRoarTime = 0;
	mPrevCamPosition = mCameraPosition;
	mLevelMap = mObjects->getMapDataPtr();
	mImmortal = false;
	mHeartBeat = 11.0f;
	mExhaust = 0.0f;

	//create mesh bases for temporary objects
	//small explosion
	mSmallExplosionBase = mObjects->createMeshBase(mObjects->generateMeshBaseUID());
	if(not mSmallExplosionBase) {
		fprintf(stderr, "IGameCtrl::init(): !mSmallExplosionBase\n");
		return false;
	}
	Triangle3Dg sTrian =
		{{Point3Dg(-5.0f, -5.0f, 0.0f), //A: x, y, z
		Point3Dg(0.0f, 0.0f, 1.0f), {0.0f, 0.0f}},    //Nx, Ny, Nz; Tx, Ty
		{Point3Dg(5.0f, 5.0f, 0.0f), //B: x, y, z
		Point3Dg(0.0f, 0.0f, 1.0f), {1.0f, 1.0f}},    //Nx, Ny, Nz; Tx, Ty
		{Point3Dg(-5.0f, 5.0f, 0.0f), //C: x, y, z
		Point3Dg(0.0f, 0.0f, 1.0f), {0.0f, 1.0f}}};   //Nx, Ny, Nz; Tx, Ty
	mSmallExplosionBase->addTrian(sTrian);
	sTrian.vertexC = sTrian.vertexB;
	sTrian.vertexB.coords.x = 5.0f;
	sTrian.vertexB.coords.y = -5.0f;
	sTrian.vertexB.tex[1] = 0.0f;
	mSmallExplosionBase->addTrian(sTrian);
	//big explosion
	mBigExplosionBase = mObjects->createMeshBase(mObjects->generateMeshBaseUID());
	if(not mBigExplosionBase) {
		fprintf(stderr, "IGameCtrl::init(): !mBigExplosionBase\n");
		return false;
	}
	sTrian.vertexA.coords.x = -30.0f;
	sTrian.vertexA.coords.y = -30.0f;
	sTrian.vertexB.coords.x = 30.0f;
	sTrian.vertexB.coords.y = 30.0f;
	sTrian.vertexC.coords.x = -30.0f;
	sTrian.vertexC.coords.y = 30.0f;
	sTrian.vertexB.tex[1] = 1.0f;
	sTrian.vertexC.tex[0] = 0.0f;
	mBigExplosionBase->addTrian(sTrian);
	sTrian.vertexC = sTrian.vertexB;
	sTrian.vertexB.coords.x = 30.0f;
	sTrian.vertexB.coords.y = -30.0f;
	sTrian.vertexB.tex[1] = 0.0f;
	mBigExplosionBase->addTrian(sTrian);

	//find a pointer to equipment base
	for(int i = 0; i < 32760; i+= 8) {
		if(mLevelMap[i+6] != 0) {
			mEquipmentBase = mObjects->mMeshes[mLevelMap[i+6]]->getBase();
			break;
		}
	}

	//gather info about animated objects
	for(MeshMap3Dg::iterator i = mObjects->mMeshes.begin(); i != mObjects->mMeshes.end(); i++) {
		Mesh3Dg *mesh = i->second;
		if(mesh->mAnimation.dwGeomAnimTime) {
			GameObject *obj = (GameObject *)mesh->getChild();
			if(obj->getType() == GOTYPE_DOOR) {
				Mesh3Dg *pFrameMesh = ((GODoor *)obj)->getFrameMesh();

				//if the door is closed
				if(pFrameMesh->mMaterialNumber == mesh->mMaterialNumber) {
					mesh->mAnimation.iElapsedTime = 0;
					pFrameMesh->setProperty(MRP_RENDERABLE, false);
				}

				if(pFrameMesh->mMaterialNumber == mesh->mMaterialNumber + 2) {
					uint32_t dwTime = mesh->mAnimation.dwGeomAnimTime / 4;
					float *pPoint = mesh->mAnimation.sTranslation.pt;
					float x = pPoint[0] * dwTime, y = pPoint[1] * dwTime, z = pPoint[2] * dwTime;
					mesh->addTranslation(x, y, z);
					mesh->mAnimation.iElapsedTime = dwTime;
				}

				if(pFrameMesh->mMaterialNumber == mesh->mMaterialNumber + 1) {
					uint32_t dwTime = mesh->mAnimation.dwGeomAnimTime;
					float *pPoint = mesh->mAnimation.sTranslation.pt;
					float x = pPoint[0] * dwTime, y = pPoint[1] * dwTime, z = pPoint[2] * dwTime;
					mesh->addTranslation(x, y, z);
					mesh->mAnimation.iElapsedTime = dwTime;
				}
			} else mesh->mAnimation.iElapsedTime = 0;
			mesh->mAnimation.iAnimStatus = ANIM_STOP;
			continue;
		}
		if(mesh->mAnimation.dwTexAnimTime) {
			mesh->mAnimation.iElapsedTime = 0;
			mesh->mAnimation.iAnimStatus = ANIM_FORWARD;
		}
	}

	//rotate sprites and find enemies
	for(GameObjectsMap::iterator i = mObjects->begin(); i != mObjects->end(); i++) {
		GameObject *obj = i->second;
		if(obj->getType() == GOTYPE_SPRITE) {
			obj->getMesh()->rotate(0.0f, -mCameraAngles.y, 0.0f);
			if(((GOSprite *)obj)->getSubType() == GOTYPE_ENEMY)
				mEnemies.push_back((GOEnemy *)obj);
		}
	}

	//if the difficulty level is EASY then remove every fourth enemy
	//and decrease agression to 4/5 of it's original value
	if(not mHighDificulty) {
		EnemiesList::size_type counter = 0;
		EnemiesList::iterator i = mEnemies.begin();
		while(i != mEnemies.end()) {
			counter++;
			if(counter % 4 == 0) {
				//hadle the iterator properly
				EnemiesList::iterator j = i;
				i++;
				//get pointers to the game object and the mesh
				GameObject *obj = *j;
				Mesh3Dg *mesh = obj->getMesh();
				//remove the enemy from the enemies list
				mEnemies.erase(j);
				//remowe the enemy as a rendered object
				mOGL->remMeshFromRenderList(mesh->getUID());
				//destroy the game object
				mObjects->destroyGameObject(obj);
				//and the mesh
				mObjects->destroyMesh(mesh->getUID());
				//reset the counter and continue
				counter = 0;
				continue;
			}
			//agression
			GOEnemy *enemy = *i;
			enemy->setAgression(enemy->getAgression() * 4 / 5);
			i++;
		}
	}

	return true;
}

void IGameCtrl::release()
{
	mEnemies.clear();
	mTempObjects.clear();
	delete this;
}

void IGameCtrl::animate()
{
	//animations
	uint32_t dwDiff = mMiliseconds;

	for(GameObjectsMap::iterator i = mObjects->begin(); i != mObjects->end(); i++) {
		GameObject *obj = i->second;
		Mesh3Dg *mesh = i->second->getMesh();
		if(obj->getType() == GOTYPE_SPRITE)
			mesh->rotate(0.0f, -mCameraAngles.y, 0.0f);

		MeshAnim3Dg *animation = &mesh->mAnimation;
		//if the object isn't beeing animated at the moment
		if(animation->iAnimStatus == ANIM_STOP)
			continue;
		//inc/dec anim time
		animation->iElapsedTime += dwDiff * animation->iAnimStatus;
		//if the object has animated texture
		if(animation->dwTexAnimTime) {
			//if texture display time went out of range then substract
			//as long as it gets back into the range
			while(animation->iElapsedTime >= (int32_t)animation->dwTexAnimTime)
				animation->iElapsedTime -= (int32_t)animation->dwTexAnimTime;
			//set next material
			mesh->mMaterialNumber = animation->dwBeginMtl +
						(animation->iElapsedTime * animation->dwNumMtls) / animation->dwTexAnimTime;
			continue;
		}

		//animate doors - if time out of range then decrease
		if(animation->iElapsedTime > (int32_t)animation->dwGeomAnimTime) {
			dwDiff -= animation->iElapsedTime - animation->dwGeomAnimTime;
			animation->iElapsedTime = animation->dwGeomAnimTime;
		} else if(animation->iElapsedTime < 0) {
			dwDiff += animation->iElapsedTime;
			animation->iElapsedTime = 0;
		}
		//calculate translation
		float *pData = animation->sTranslation.pt;
		float x = pData[0] * dwDiff * animation->iAnimStatus;
		float y = pData[1] * dwDiff * animation->iAnimStatus;
		float z = pData[2] * dwDiff * animation->iAnimStatus;
		mesh->addTranslation(x, y, z);
		//change door frame texture if anim finished
		uint32_t dwElTime = animation->iElapsedTime;
		if(obj->getType() != GOTYPE_DOOR)
			continue;
		if(dwElTime == animation->dwGeomAnimTime)
			((GODoor *)obj)->getFrameMesh()->mMaterialNumber = mesh->mMaterialNumber + 1;
		//if door are closed then don't draw it's frame (it is not necesary)
		else if(dwElTime == 0) {
			((GODoor *)obj)->getFrameMesh()->mMaterialNumber = mesh->mMaterialNumber;
			((GODoor *)obj)->getFrameMesh()->setProperty(MRP_RENDERABLE, false);
		}
		else continue;
		animation->iAnimStatus = ANIM_STOP;
		dwDiff = mMiliseconds;
	}
	//temporary objects
	TempObjList::iterator i = mTempObjects.begin();
	while(i != mTempObjects.end()) {
		TemporaryObject *obj = *i;
		//moving objects
		if(obj->getRange() > 0.0f) {
			Weapon *weapon = (Weapon *)obj->getUserData();
			float speed = 0.2f;
			if(weapon->getType() == UIT_FLAMER)
				speed = 0.15f;
			float dist = mMiliseconds*speed*mConfig->getGameSpeed();
			Mesh3Dg *mesh = obj->getMesh();
			Vector3Dg dir = *obj->getVelocityDirection();
			if(bulletCollision(weapon, dist, mesh->geomCenter(), dir, obj->getOrigin()) or not obj->update(mMiliseconds, dist)) {
				TempObjList::iterator j = i;
				i++;
				mTempObjects.erase(j);
				mOGL->remMeshFromRenderList(mesh->getUID());
				mObjects->destroyGameObject(obj);
				mObjects->destroyMesh(mesh->getUID());
			} else {
				float x = dir.x * dist;
				float z = dir.z * dist;
				mesh->addTranslation(x, 0.0f, z);
				i++;
			}
		//static objects
		} else if(obj->getLifeTime() > 0) {
			if(not obj->update(mMiliseconds)) {
				TempObjList::iterator j = i;
				i++;
				mTempObjects.erase(j);
				Mesh3Dg *mesh = obj->getMesh();
				mOGL->remMeshFromRenderList(mesh->getUID());
				mObjects->destroyGameObject(obj);
				mObjects->destroyMesh(mesh->getUID());
			}
			else i++;
		}
	}
}

bool IGameCtrl::update(UpdateData *pData)
{
	mVelocity.x = *(pData->pX) - mPrevCamPosition.x;
	mVelocity.z = *(pData->pZ) - mPrevCamPosition.z;
	mLostEnergy = 0;
	//inertia simulation (after bumping walls, beeing shot, etc.)
	if(mAuxVTime > 0) {
		mVelocity = mAuxiliaryV;
		if(mMiliseconds) {
			mAuxVTime -= (int32_t)mMiliseconds;
			//inertia should last 250ms
			float fDissipation = (0.004 * (float)mMiliseconds);
			if(fDissipation < 0.0f)
				fDissipation = 0.0f;
			//dissipation
			mAuxiliaryV *= 1.0f - fDissipation;
		}
	}
	if(mAuxVTime < 0) {
		mAuxiliaryV.x = 0.0f;
		mAuxiliaryV.z = 0.0f;
		mAuxVTime = 0;
	}

	//calm down (if moving than calm down slower)
	if(mVelocity.x or mVelocity.z)
		mHeartBeat -= 0.004f * mMiliseconds;
	else
		mHeartBeat -= 0.008f * mMiliseconds;
	if(mHeartBeat < 11.0f)
		mHeartBeat = 11.0f;
	//set heart beat speed in OSD
	mOGL->mOSD.setHeartSpeed(mHeartBeat);
	mExhaust = (mHeartBeat - 11.0f) / 45.0f;

	mCameraAngles.y = pData->fRotY;
	mCameraDirection.x = pData->fDirX;
	mCameraDirection.z = pData->fDirZ;
	mSinCameraRotation = sinf(mCameraAngles.y);
	mCosCameraRotation = cosf(mCameraAngles.y);
	//offset to field on which is camera
	uint32_t dwPosOffset = mPosition[0] * 8 + mPosition[1] * 512;
	//first collisions with walls, door and edges
	bool bEdges[] = {false, false, false, false, false};
	//get info about existing walls in actual square
	uint16_t wWalls = mLevelMap[dwPosOffset + 4];
	for(uint32_t i = 0; i < 4; ++i) {
		if(wWalls & (1 << i)) {
			Mesh3DgUID dwMeshNum = mLevelMap[dwPosOffset + i];
			bEdges[i] = bEdges[i + 1] = true;
			GameObject *obj = mObjects->getGameObject(dwMeshNum);
			if(obj->getPlayerCollisionType() != PCTYPE_BUMP)
				continue;
			switch(obj->getType()) {
				case GOTYPE_WALL:
					mLostEnergy += wallCollision(obj, i);
					break;
				case GOTYPE_DOOR: {
					//only if not open
					if(isDoorOpen((GODoor *)obj))
						mLostEnergy += wallCollision(obj, i);
					break;
				}
				default:
					break;
			}
		}
	}
	bEdges[0] = bEdges[0] or bEdges[4];
	//search for edges of walls from neighbour squares
	searchForEdges(bEdges, mPosition[0], mPosition[1]);
	//check edges collisions
	edgesCollision(bEdges);

	//actualize map
	actualizeMap(mPosition[0], mPosition[1], wWalls);

	//check for equipment
	if(mLevelMap[dwPosOffset + 4] & 32) {
		Mesh3Dg *mesh = mObjects->mMeshes[mLevelMap[dwPosOffset + 6]];
		if(equipmentCollision(mesh, dwPosOffset)) {
			pData->iTakenItem = mesh->mMaterialNumber;
			//remove info from map
			mLevelMap[dwPosOffset + 4] &= ~32;
			//don't render
			mesh->setProperty(MRP_RENDERABLE, false);
		}
	}
	//check for sth that may block the way
	if(mLevelMap[dwPosOffset + 4] & 16) {
		Mesh3DgUID dwMeshNum = mLevelMap[dwPosOffset + 5];
		Mesh3Dg *mesh = mObjects->mMeshes[dwMeshNum];
		if(((GameObject *)mesh->getChild())->getPlayerCollisionType() == PCTYPE_STOP)
			spriteCollision(mesh);
	}
	//enemy also blocks the way (and a burning enemy can burn the player,
	//so he can lost energy)
	mLostEnergy += enemyCollision();

	//actualize position
	mCameraPosition += mVelocity;

	//close doors from previous field (if any)
	uint16_t wPrevOffset = (uint16_t)(mPrevPosition[0] * 8 + mPrevPosition[1] * 512);
	if(not (mLevelMap[wPrevOffset + 4] & 64) and wPrevOffset != dwPosOffset)
		closePrevDoor(wPrevOffset, dwPosOffset);
	//open unlocked doors from actualposition
	for(uint32_t i = 0; i < 4; ++i) {
		if(not (wWalls & (1 << i)))
			continue;
		Mesh3DgUID num = mLevelMap[dwPosOffset + i];
		GameObject *obj = mObjects->getGameObject(num);
		//if it is door and it isn't locked
		if((obj->getType() == GOTYPE_DOOR) and not (mLevelMap[dwPosOffset + 4] & 64))
			openDoor(num, (GODoor *)obj);
	}

	mPrevPosition[0] = mPosition[0];
	mPrevPosition[1] = mPosition[1];
	mPosition[0] = (uint8_t)(*(pData->pX) / 100.0f);
	mPosition[1] = (uint8_t)-(*(pData->pZ) / 100.0f);
	dwPosOffset = mPosition[0] * 8 + mPosition[1] * 512;
	mMiliseconds = pData->uiMiliseconds;

	//do animations
	animate();

	//check if teleport collision
	if(mLevelMap[dwPosOffset + 4] & 16) {
		Mesh3Dg *mesh = mObjects->mMeshes[mLevelMap[dwPosOffset + 5]];
		if(((GameObject *)mesh->getChild())->getPlayerCollisionType() == PCTYPE_TELEPORT) {
			uint32_t uiPastTime = 0;
			//if teleportCollision() returned 'false' then a collision with
			//an ending teleport ocured
			if(not teleportCollision(mesh, dwPosOffset, uiPastTime))
				return false;
			pData->uiTeleportationTime = uiPastTime;
		}
	}

	*(pData->pX) = mPrevCamPosition.x = mCameraPosition.x;
	*(pData->pZ) = mPrevCamPosition.z = mCameraPosition.z;
	pData->pPosition[0] = mPosition[0];
	pData->pPosition[1] = mPosition[1];

	//update enemies
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++) {
		GOEnemy *enemy = *i;
		GOEnemyUpdateResult result = enemy->update(mMiliseconds, mCameraPosition);
		switch(result) {
			case EUR_AIMING: {
				Vector3Dg vec = mCameraPosition - enemy->getMesh()->geomCenter();
				float distaince = vec.length();
				//find apropriate weapon in map
				UsableItemType type = UsableItemType(enemy->getWeaponType());
				WeaponsMap::iterator iter = mEnemyWeapons.find(type);
				if(iter == mEnemyWeapons.end())
					break;
				if(distaince > iter->second.getRange())
					break;
				if(canShoot(distaince, enemy->getMesh()->geomCenter(), enemy->getDirection()))
					enemy->shoot();
				break;
			}
			case EUR_SHOT: {
				Vector3Dg pos = enemy->getMesh()->geomCenter();
				mSDL->playSound(mWeaponSounds[enemy->getWeaponType()-1], 3,
								calculateVolume(pos, VOLDIST_TYPE2));
				//find apropriate weapon in map
				UsableItemType type = UsableItemType(enemy->getWeaponType());
				WeaponsMap::iterator iter = mEnemyWeapons.find(type);
				if(iter == mEnemyWeapons.end())
					break;
				Weapon *weapon = &(iter->second);
				switch(type) {
					case UIT_HANDGUN:
					case UIT_SHOTGUN:
					case UIT_MASHINEGUN: {
						for(uint16_t i = 0; i < weapon->getBullets(); i++)
							ammoCollision(weapon, pos, enemy->getDirection());
						break;
					}
					case UIT_FLAMER:
					case UIT_BLASTER:
					case UIT_ROCKETLAUNCHER: {
						pos = pos + (enemy->getDirection() * 25.0f);
						pos.x += mCosCameraRotation * enemy->getAmmoXPosShift();
						pos.z += mSinCameraRotation * enemy->getAmmoXPosShift();
						pos.y = enemy->getAmmoYPosShift();
						Vector3Dg dir = mCameraPosition - pos;
						dir.y = 0.0f;
						dir.normalize();
						fireAmmo(weapon, pos, dir);
						break;
					}
					default:
						break;
				}
				break;
			}
			case EUR_WALK: {
				//random roar
				if(mLastRoarTime > 100) {
					randomRoar(enemy);
					mLastRoarTime = 0;
				}
				//NOTE: don't break here!
			}
			case EUR_BURNING: {
				checkEnemyCollisions(enemy);
				enemy->move();
				break;
			}

			default:
				break;
		}
		mLastRoarTime += mMiliseconds;
	}

	pData->uiLostEnergy = mLostEnergy;

	//decrease mashinegun next firing delay
	if(mMGFireDiff < MACHINEGUN_FIRE_DELAY)
		mMGFireDiff += mMiliseconds;
	//the same for payer burn delay
	if(mPlayerBurnDiff < PLAYER_BURN_DELAY)
		mPlayerBurnDiff += mMiliseconds;
	return true;
}

void IGameCtrl::searchForEdges(bool *edges, uint8_t posX, uint8_t posZ)
{
	//search for edges of walls from neighbour squares
	uint16_t posWInfoOffs = (uint16_t)(getRelativePosOffset(posX, posZ, -1, 0) + 4);
	uint16_t posEInfoOffs = (uint16_t)(getRelativePosOffset(posX, posZ, 1, 0) + 4);
	uint16_t posSInfoOffs = (uint16_t)(getRelativePosOffset(posX, posZ, 0, -1) + 4);
	uint16_t posNInfoOffs = (uint16_t)(getRelativePosOffset(posX, posZ, 0, 1) + 4);
	uint16_t xm1 = mLevelMap[posWInfoOffs];
	uint16_t xp1 = mLevelMap[posEInfoOffs];
	uint16_t zm1 = mLevelMap[posSInfoOffs];
	uint16_t zp1 = mLevelMap[posNInfoOffs];
	if((xm1 & 1) or (zp1 & 8)) edges[0] = true;
	if((xp1 & 1) or (zp1 & 2)) edges[1] = true;
	if((xp1 & 4) or (zm1 & 2)) edges[2] = true;
	if((xm1 & 4) or (zm1 & 8)) edges[3] = true;
}

uint16_t IGameCtrl::getRelativePosOffset(int8_t posX, int8_t posZ, int8_t relX, int8_t relZ)
{
	int8_t x = posX + relX;
	if(x > 64)
		x -= 64;
	if(x < 0)
		x += 64;
	int8_t z = posZ + relZ;
	if(z > 64)
		z -= 64;
	if(z < 0)
		z += 64;
	return (uint16_t)(x * 8 + z * 512);
}

void IGameCtrl::revealMap()
{
	for(int32_t i = 0; i < 0x8000; i += 8)
		mLevelMap[i + 4] |= 128;
}

void IGameCtrl::actualizeMap(int8_t x, int8_t z, uint16_t walls)
{
	//uncover actual position
	mLevelMap[(x * 8 + z * 512) + 4] |= 128;
	//and other around
	uint16_t posWInfoOffs = (uint16_t)(getRelativePosOffset(x, z, -1, 0) + 4);
	uint16_t posEInfoOffs = (uint16_t)(getRelativePosOffset(x, z, 1, 0) + 4);
	uint16_t posSInfoOffs = (uint16_t)(getRelativePosOffset(x, z, 0, -1) + 4);
	uint16_t posNInfoOffs = (uint16_t)(getRelativePosOffset(x, z, 0, 1) + 4);
	uint16_t xm1 = mLevelMap[posWInfoOffs];
	uint16_t xp1 = mLevelMap[posEInfoOffs];
	uint16_t zm1 = mLevelMap[posSInfoOffs];
	uint16_t zp1 = mLevelMap[posNInfoOffs];
	//if these is no northern wall
	if(not (walls & 1)) {
		//reveal position to the north and, if necessary, left and right from there
		mLevelMap[posNInfoOffs] |= 128;
		if(not (zp1 & 8)) mLevelMap[getRelativePosOffset(x, z, -1, 1) + 4] |= 128;
		if(not (zp1 & 2)) mLevelMap[getRelativePosOffset(x, z, 1, 1) + 4] |= 128;
	}
	//eastern
	if(not (walls & 2)) {
		mLevelMap[posEInfoOffs] |= 128;
		if(not (xp1 & 1)) mLevelMap[getRelativePosOffset(x, z, 1, 1) + 4] |= 128;
		if(not (xp1 & 4)) mLevelMap[getRelativePosOffset(x, z, 1, -1) + 4] |= 128;
	}
	//southern
	if(not (walls & 4)) {
		mLevelMap[posSInfoOffs] |= 128;
		if(not (zm1 & 8)) mLevelMap[getRelativePosOffset(x, z, -1, -1) + 4] |= 128;
		if(not (zm1 & 2)) mLevelMap[getRelativePosOffset(x, z, 1, -1) + 4] |= 128;
	}
	//western
	if(not (walls & 8)) {
		mLevelMap[posWInfoOffs] |= 128;
		if(not (xm1 & 1)) mLevelMap[getRelativePosOffset(x, z, -1, 1) + 4] |= 128;
		if(not (xm1 & 4)) mLevelMap[getRelativePosOffset(x, z, -1, -1) + 4] |= 128;
	}
}

void IGameCtrl::redrawMap()
{
	//redraw map
	for(int32_t i = 0; i < 0x8000; i += 8) {
		if(not (mLevelMap[i + 4] & 128))
			continue;
		uint8_t btZ = (uint8_t)(i / 512);
		uint8_t btX = (uint8_t)((i % 512) / 8);
		mOGL->mOSD.drawMapPos(&mLevelMap[i], btX, btZ, mObjects);
	}
}

uint32_t IGameCtrl::getFaceDir()
{
	float fRotY = mCameraAngles.y;
	//north
	if(fRotY > ANG_315DEG or fRotY <= ANG_45DEG)
		return 0;
	//east
	if(fRotY > ANG_45DEG and fRotY <= ANG_135DEG)
		return 1;
	//south
	if(fRotY > ANG_135DEG and fRotY <= ANG_225DEG)
		return 2;
	//west - no need to check - only one possibility left
	//if(fRotY > ANG_225DEG and fRotY <= ANG_315DEG)
	return 3;
}

void IGameCtrl::handCollision(UsableItem *pHand)
{
	uint32_t dwPosOffset = mPosition[0] * 8 + mPosition[1] * 512;
	uint32_t dwFaceDir = getFaceDir();
	uint32_t dwObjOffset = dwPosOffset + dwFaceDir;
	Mesh3DgUID dwMeshNum = mLevelMap[dwObjOffset];
	//if these's a wall
	if((mLevelMap[dwPosOffset + 4] & (1 << dwFaceDir))) {
		MapDataPtr pData = &mLevelMap[0x8000];
		uint32_t dwDataOffset = getDataOffset(dwObjOffset, pData);
		int32_t iMtlNum = mObjects->mMeshes[dwMeshNum]->mMaterialNumber;
		switch(iMtlNum) {
			//if button, and it is switched-off - switch it on
			case 21:
			case 104:
			case 121:
			case 138: {
				//change material
				mObjects->mMeshes[dwMeshNum]->mMaterialNumber++;
				//display text
				mOGL->mOSD.addText(mTexts[BUTTON_TEXTS].getText());
				mSDL->playSound(pHand->mUseSoundNumber, 1);
				//if action desc not found
				if(dwDataOffset == 0)
					return;
				dwDataOffset++;PRE_LOOP();
				//do the action
				while(pData[dwDataOffset] < 0xFFFE) {
					doActions(dwDataOffset, pData);
					dwDataOffset += 2;
				}AFTER_LOOP();
				return;
			}
			//if switched-on button
			case 22:
			case 105:
			case 122:
			case 139: {
				//change material
				mObjects->mMeshes[dwMeshNum]->mMaterialNumber--;
				//display text
				mOGL->mOSD.addText(mTexts[BUTTON_TEXTS].getText());
				mSDL->playSound(pHand->mUseSoundNumber, 1);
				//if action desc not found
				if(dwDataOffset == 0)
					return;
				//search for 'switch off button' action
PRE_LOOP();				while(pData[dwDataOffset] != 0xFFFE) {
					dwDataOffset++;
					//if we got to next object action desc than this button
					//does not have switch off action
					if(pData[dwDataOffset] == 0xFFFF){AFTER_LOOP();
						return;}
				}AFTER_LOOP();
				//do the action
				dwDataOffset++;PRE_LOOP();
				while(pData[dwDataOffset] != 0xFFFF) {
					doActions(dwDataOffset, pData);
					dwDataOffset += 2;
				}AFTER_LOOP();
				return;
			}
			//display apropirate texts
			//wall with no collision
			case 26:
				mOGL->mOSD.addText(mTexts[HAND_NOTHING_TEXTS].getText());
				return;
			//if door
			case 14:
			case 17: {
				GameObject *obj = mObjects->getGameObject(dwMeshNum);
				if(obj->getType() != GOTYPE_DOOR)
					return;
				//get frame's material number
				Mesh3Dg *pMesh = ((GODoor *)obj)->getFrameMesh();
				iMtlNum = pMesh->mMaterialNumber;
				switch(iMtlNum) {
					//closed
					case 14:
					case 17:
						mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_CLOSED_DOOR));
						return;
					//opening
					case 16:
					case 19:
						mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_OPENING_DOOR));
						return;
					//open
					case 15:
					case 18:
						mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_OPEN_DOOR));
						return;
				}
				return;
			}
			//wall
			default: {
				GameObject *obj = mObjects->getGameObject(dwMeshNum);
				if(obj->getType() != GOTYPE_WALL)
					return;
				//with slot
				Mesh3Dg *slot = ((GOWall *)obj)->getSlotMesh();
				if(slot) {
					switch(slot->mMaterialNumber) {
						case 51:
						case 142:
						case 145:
						case 148:
							mOGL->mOSD.addText(mTexts[HAND_SLOT_TEXTS].getText());
							return;
						case 52:
						case 143:
						case 146:
						case 149:
							mOGL->mOSD.addText(mTexts[HAND_SLOT_USED_TEXTS].getText());
							return;
						default:
							break;
					}
				}
				//with blood
				if(mObjects->mMaterials[obj->getMesh()->mMaterialNumber]->uiTexturesQuantity == 2) {
					mOGL->mOSD.addText(mTexts[HAND_BLOOD_TEXTS].getText());
					return;
				}
				//normal wall
				mOGL->mOSD.addText(mTexts[HAND_WALL_TEXTS].getText());
				return;
			}
		}
		return;
	}

	//enemy
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++) {
		GOEnemy *enemy = *i;
		Vector3Dg pos = enemy->getMesh()->geomCenter();
		if(((int)pos.x)/100 == mPosition[0] and ((int)-pos.z)/100 == mPosition[1]) {
			if(enemy->isAlive())
				mOGL->mOSD.addText(mTexts[HAND_ENEMY_TEXTS].getText());
			else mOGL->mOSD.addText(mTexts[HAND_CORPS_TEXTS].getText());
			return;
		}
	}
	//equipment
	if(mLevelMap[dwPosOffset + 4] & 32) {
		mOGL->mOSD.addText(mTexts[HAND_EQUIPMENT_TEXTS].getText());
		return;
	}
	//corps
	if(mLevelMap[dwPosOffset + 7]) {
		mOGL->mOSD.addText(mTexts[HAND_CORPS_TEXTS].getText());
		return;
	}
	//sprite/blocade
	if(mLevelMap[dwPosOffset + 4] & 16) {
		dwMeshNum = mLevelMap[dwPosOffset + 5];
		int32_t iMtlNum = mObjects->mMeshes[dwMeshNum]->mMaterialNumber;
		if(iMtlNum == 27 or iMtlNum == 28)
			mOGL->mOSD.addText(mTexts[HAND_BLOCADE_TEXTS].getText());
		else
			mOGL->mOSD.addText(mTexts[HAND_SPRITE_TEXTS].getText());
		return;
	}
	mOGL->mOSD.addText(mTexts[HAND_NOTHING_TEXTS].getText());
}

bool IGameCtrl::isDoorOpen(GODoor *door)
{
	Mesh3Dg *pMesh = door->getMesh();
	if(pMesh->mMaterialNumber + 1 != door->getFrameMesh()->mMaterialNumber)
		return true;
	return false;
}

float IGameCtrl::wallCollisionAmmo(Vector3Dg &begin, Vector3Dg &end, uint8_t xpos, uint8_t zpos,
									uint32_t walldir, GameObject *obj)
{
	//copy the coordinates of the first vertex in the first triangle
	Point3Dg pt = *obj->getMesh()->getBase()->getVCoordsStream();
	switch(walldir) {
		case 1:
		case 3: {
			pt.z = (pt.x - begin.x) * ((end.z - begin.z) / (end.x - begin.x)) + begin.z;
			float max = -(float)(zpos * 100.0f), min = -(float)((zpos+1) * 100.0f);
			//if the coordinate is not on the wall
			if(pt.z < min or pt.z > max)
				return 0.0f;
		}
		break;
		case 0:
		case 2: {
			pt.x = (pt.z - begin.z) * ((end.x - begin.x) / (end.z - begin.z)) + begin.x;
			float min = xpos * 100.0f, max = (xpos+1) * 100.0f;
			if(pt.x < min or pt.x > max)
				return 0.0f;
		}
		break;
	}
	Vector3Dg vec(pt.x - begin.x, 0.0f, pt.z - begin.z);
	return vec.length();
}

float IGameCtrl::spriteCollisionAmmo(Vector3Dg &begin, Vector3Dg &end, uint8_t xpos, uint8_t zpos, float range)
{
	Vector3Dg spritepos(xpos * 100.0f + 50.0f, begin.y, -(float)(zpos * 100.0f + 50.0f));
	Vector3Dg ammoway = end - begin;
	//a vector from 'begin point' to 'sprite pos'
	Vector3Dg vec = spritepos - begin;
	float dotprod = vec.dotProduct(ammoway);
	//if the sprite is behind the player
	if(dotprod < 0.0f)
		return 0.0f;
	//calculate the distaince from the sprite's plane
	float spritedist = dotprod / range;
	//if the sprite is out of ammo's range
	if(spritedist > range)
		return 0.0f;
	//calculate the intersection point
	ammoway *= spritedist / range;
	vec = begin + ammoway;
	//calculate the distaince between the intersection point and the sprite's position
	vec -= spritepos;
	if(vec.length() > 25.0f)
		return 0.0f;
	return ammoway.length();
}

float IGameCtrl::enemyCollisionAmmo(Vector3Dg &begin, Vector3Dg &end, Vector3Dg &enemypos, float range)
{
	Vector3Dg ammoway = end - begin;
	ammoway.y = 0.0f;
	//a vector from 'begin point' to 'sprite pos'
	Vector3Dg vec = enemypos - begin;
	vec.y = 0.0f;
	float dotprod = vec.dotProduct(ammoway);
	//if the elemy is behind the player
	if(dotprod < 0.0f)
		return 0.0f;
	//calculate the distaince from the sprite's plane
	float enemydist = dotprod / range;
	//if the sprite is out of ammo's range
	if(enemydist > range)
		return 0.0f;
	//calculate the intersection point
	ammoway *= enemydist / range;
	vec = begin + ammoway;
	//calculate the distaince between the intersection point and the sprite's position
	vec -= enemypos;
	vec.y = 0.0f;
	if(vec.length() > 25.0f)
		return 0.0f;
	return ammoway.length();
}

float IGameCtrl::playerCollisionAmmo(Vector3Dg &begin, Vector3Dg &end, float range)
{
	Vector3Dg ammoway = end - begin;
	//a vector from 'begin point' to 'sprite pos'
	Vector3Dg vec = mCameraPosition - begin;
	vec.y = 0.0f;
	float dotprod = vec.dotProduct(ammoway);
	//if the player is behind the shooter
	if(dotprod < 0.0f)
		return 0.0f;
	//calculate the distaince from the player
	float dist = dotprod / range;
	//if the player is out of ammo's range
	if(dist > range)
		return 0.0f;
	//calculate the intersection point
	ammoway *= dist / range;
	vec = begin + ammoway;
	vec.y = 0.0f;
	//calculate the distaince between the intersection point and the player's position
	vec -= mCameraPosition;
	if(vec.length() > mCamRay)
		return 0.0f;
	return ammoway.length();
}

void IGameCtrl::fireAmmo(Weapon *weapon, Vector3Dg position, Vector3Dg direction)
{
	Mesh3Dg *mesh = mObjects->createMesh(mObjects->generateMeshUID());
	if(mesh == 0)
		return;
	TemporaryObject *obj = mObjects->createTempGameObject(mesh, weapon->getRange(), direction);
	if(obj == 0) {
		mObjects->destroyMesh(mesh->getUID());
		return;
	}
	//user data
	obj->setUserData(weapon);
	//render properties for mesh
	mesh->setProperty(MRP_TRANSPARENT, true);
	mesh->setProperty(MRP_Z_MINUS_OFFSET, true);
	//position and rotation
	mesh->setPosition(position.x, 50.0f + position.y, position.z);
	mesh->rotate(0.0f, -mCameraAngles.y, 0.0f);
	//origin of the object
	obj->updateOrigin();

	switch(weapon->getType()) {
		case UIT_FLAMER: {
			//mesh base
			mesh->setBase(mBigExplosionBase);
			//animation
			mesh->mAnimation.dwTexAnimTime = 400;
			mesh->mAnimation.dwBeginMtl = 75;
			mesh->mAnimation.dwNumMtls = 2;
			mesh->mMaterialNumber = mesh->mAnimation.dwBeginMtl;
			break;
		}
		case UIT_BLASTER: {
			//mesh base
			mesh->setBase(mSmallExplosionBase);
			//animation
			mesh->mAnimation.dwTexAnimTime = 400;
			mesh->mAnimation.dwBeginMtl = 77;
			mesh->mAnimation.dwNumMtls = 2;
			mesh->mMaterialNumber = mesh->mAnimation.dwBeginMtl;
			break;
		}
		case UIT_ROCKETLAUNCHER: {
			//mesh base
			mesh->setBase(mSmallExplosionBase);
			//animation
			mesh->mAnimation.dwTexAnimTime = 400;
			mesh->mAnimation.dwBeginMtl = 79;
			mesh->mAnimation.dwNumMtls = 2;
			mesh->mMaterialNumber = mesh->mAnimation.dwBeginMtl;
			break;
		}
		default: {
			mObjects->destroyGameObject(obj);
			mObjects->destroyMesh(mesh->getUID());
			return;
		}
	}
	//add the object to the render and temp-objects lists
	mTempObjects.push_back(obj);
	mOGL->addMeshToRenderList(mesh);
}

bool IGameCtrl::bulletCollision(Weapon *weapon, float range, Vector3Dg position, Vector3Dg direction, Vector3Dg &origin)
{
	//calculate the end point of the ammo's way
	Vector3Dg end = position + (direction * range);
	//calculate the boundaries of potentially affected area
	int8_t beginx = (int8_t)(position.x / 100), beginz = (int8_t)(-position.z / 100);
	int8_t endx = (int8_t)(end.x / 100), endz = (int8_t)(-end.z / 100);

	//if out of bounds
	if(beginx < 0 or beginx > 63 or beginz < 0 or beginz > 63 or
		endx < 0 or endx > 63 or endz < 0 or endz > 63)
		return false;

	//position's offset in the map buffer
	uint32_t offset = beginx * 8 + beginz * 512;
	//pointer to the closest object on the bullet's way
	GameObject *closestobject = 0;
	float distaince = range;

	//if the bullet's way doesn't intersect any wall's plane
	if(beginx == endx and beginz == endz)
		goto SpriteCollisionCheck;

	//which walls to check
	{
		uint32_t count = 0, check[2];
		//north
		if(direction.z < 0.0f) {
			check[count] = 0;
			count++;
		}
		//south
		else if(direction.z > 0.0f) {
			check[count] = 2;
			count++;
		}
		//east
		if(direction.x > 0.0f) {
			check[count] = 1;
			count++;
		}
		//west
		else if(direction.x < 0.0f) {
			check[count] = 3;
			count++;
		}

		bool edges[] = {false, false, false, false, false};
		//position's offset in the map buffer
		uint16_t walls = mLevelMap[offset + 4];
		for(uint32_t i = 0; i < count; ++i) {
			if(not (walls & (1 << check[i])))
				continue;
			Mesh3DgUID meshnum = mLevelMap[offset + check[i]];
			edges[check[i]] = edges[check[i] + 1] = true;
			GameObject *obj = mObjects->getGameObject(meshnum);
			if(obj->getAmmoCollisionType() == ACTYPE_NONE)
				continue;
			float dist = 0.0f;
			switch(obj->getType()) {
				case GOTYPE_WALL:
					dist = wallCollisionAmmo(position, end, beginx, beginz, check[i], obj);
					break;
				case GOTYPE_DOOR: {
					//only if not open
					if(isDoorOpen((GODoor *)obj))
						dist = wallCollisionAmmo(position, end, beginx, beginz, check[i], obj);
					break;
				}
				default:
					break;
			}
			if(dist != 0.0f and dist < distaince) {
				distaince = dist;
				closestobject = obj;
			}
		}
		edges[0] = edges[0] or edges[4];
		searchForEdges(edges, beginx, beginz);
		//check edges collisions
		Vector3Dg explosionpos;
		if(bulletEdgesCollision(edges, position, end, beginx, beginz, 4.0f, explosionpos)) {
			explosionpos.y = position.y;
			switch(weapon->getType()) {
				case UIT_FLAMER:
					mSDL->playSound(mCollisionSounds[CSND_EXPLOSION], 3, calculateVolume(explosionpos, VOLDIST_TYPE1));
					displayExplosion(explosionpos, EXPLT_BIG, 0, false);
				break;
				case UIT_BLASTER:
					displayExplosion(explosionpos, EXPLT_ELECTRIC, 0, false);
				break;
				case UIT_ROCKETLAUNCHER:
					mSDL->playSound(mCollisionSounds[CSND_EXPLOSION], 3, calculateVolume(explosionpos, VOLDIST_TYPE1));
					displayHugeExplosion(explosionpos, direction);
					explosionCollision(explosionpos);
				break;
				default:
					return false;
			}
			return true;
		}
	}

	SpriteCollisionCheck:
	//sprites
	Mesh3DgUID meshnum = mLevelMap[offset + 5];
	if(mLevelMap[offset + 4] & 16) {
		GameObject *obj = mObjects->getGameObject(meshnum);
		if(obj->getAmmoCollisionType() != ACTYPE_NONE) {
			float dist = spriteCollisionAmmo(position, end, beginx, beginz, range);
			if(dist != 0.0f and dist < distaince) {
				distaince = dist;
				closestobject = obj;
			}
		}
	}

	//now check for enemies
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++) {
		GameObject *obj = (GameObject *)*i;
		if(obj->getAmmoCollisionType() != ACTYPE_NONE) {
			Vector3Dg enemypos = obj->getMesh()->geomCenter();
			float dist = enemyCollisionAmmo(position, end, enemypos, weapon->getRange());
			if(dist != 0.0f and dist < distaince) {
				distaince = dist;
				closestobject = obj;
			}
		}
	}

	//check the player
	float playerDistaince = playerCollisionAmmo(position, end, weapon->getRange());
	//if the distaince to the player is less than the distaince to the closest object
	//then the player is hit
	if(playerDistaince > 0.0 and playerDistaince < distaince) {
		playerHit(weapon->getPower(), direction);
		return true;
	}

	//if any object has been hit
	if(closestobject) {
		Vector3Dg explosionpos = position + (direction * (distaince - 5.0f));
		switch(closestobject->getAmmoCollisionType()) {
			case ACTYPE_UNBREAKABLE: {
				switch(weapon->getType()) {
					case UIT_FLAMER:
						mSDL->playSound(mCollisionSounds[CSND_EXPLOSION], 3, calculateVolume(explosionpos, VOLDIST_TYPE1));
						displayExplosion(explosionpos, EXPLT_BIG, 0, false);
					break;
					case UIT_BLASTER:
						displayExplosion(explosionpos, EXPLT_ELECTRIC, 0, false);
					break;
					case UIT_ROCKETLAUNCHER:
						mSDL->playSound(mCollisionSounds[CSND_EXPLOSION], 3, calculateVolume(explosionpos, VOLDIST_TYPE1));
						displayHugeExplosion(explosionpos, direction);
						explosionCollision(explosionpos);
					break;
					default:
						return false;
				}
				return true;
			}
			case ACTYPE_EXPLODABLE: {
				if(closestobject->getType() != GOTYPE_SPRITE)
					return false;
				//destroy the object if the weapon was not a blaster
				if(weapon->getType() == UIT_BLASTER)
					displayExplosion(explosionpos, EXPLT_ELECTRIC, 0, false);
				else {
					closestobject->getMesh()->setProperty(MRP_RENDERABLE, false);
					//rem info in map
					mLevelMap[offset + 4] &= ~16;
					//if there is an equipment near the explodable object then destroy it to
					if(mLevelMap[offset + 4] & 32) {
						mObjects->mMeshes[mLevelMap[offset + 6]]->setProperty(MRP_RENDERABLE, false);
						//rem info in map
						mLevelMap[offset + 4] &= ~32;
					}
					displayHugeExplosion(explosionpos, direction);
					explosionCollision(explosionpos);
					mSDL->playSound(mCollisionSounds[CSND_EXPLOSION], 3, calculateVolume(explosionpos, VOLDIST_TYPE1));
				}
				return true;
			}
			case ACTYPE_ENEMY: {
				GOEnemy *enemy = (GOEnemy *)closestobject;
				//decrease the HP and die, if necessary
				//(yes, in the original game, the enemy died after his HP went below 0)
				if(not enemy->isBurning()) {
					switch(weapon->getType()) {
						case UIT_FLAMER: {
							if(enemy->burn()) {
								//turn off other fire sounds
								unregisterFireSounds();
								mSDL->playSound(mEnemySounds[ESND_FIRE], 5, 128, -1);
								enemy->registerSound(5);
							} else hitRoar(enemy);
							break;
						}
						case UIT_BLASTER: {
							if(not enemy->canBeParalized() or not enemy->isAlive())
								break;
							if(enemy->hit(weapon->getPower(), mCameraPosition) < 0) {
								enemy->die();
								//sometimes leave a present
								leaveAmmo(enemy);
								//blood
								Vector3Dg way = explosionpos - origin;
								spreadBlood(enemy, direction, way.length());

								mSDL->playSound(mEnemySounds[ESND_DEATH], 4, calculateVolume(explosionpos, VOLDIST_TYPE3));
							} else hitRoar(enemy);
							checkEnemyCollisions(enemy);
							enemy->move();
							break;
						}
						case UIT_ROCKETLAUNCHER: {
							if(enemy->isAlive()) {
								if(enemy->hit(weapon->getPower(), mCameraPosition) < 0) {
									enemy->die();
									//sometimes leave a present
									leaveAmmo(enemy);
									//blood
									Vector3Dg way = explosionpos - origin;
									spreadBlood(enemy, direction, way.length());
	
									mSDL->playSound(mEnemySounds[ESND_DEATH], 4, calculateVolume(explosionpos, VOLDIST_TYPE3));
								}
								checkEnemyCollisions(enemy);
								enemy->move();
							}
							break;
						}
						default:
							break;
					}
				}

				//explosion and sound
				if(weapon->getType() == UIT_BLASTER)
					displayExplosion(explosionpos, EXPLT_ELECTRIC, 0, false);
				else {
					if(weapon->getType() == UIT_ROCKETLAUNCHER) {
						displayHugeExplosion(explosionpos, direction);
						explosionCollision(explosionpos);
					} else displayExplosion(explosionpos, EXPLT_BIG, 0, false);
					mSDL->playSound(mCollisionSounds[CSND_EXPLOSION], 3, calculateVolume(explosionpos, VOLDIST_TYPE1));
				}
				return true;
			}
			default:
				return false;
		}
	}
	return false;
}

void IGameCtrl::ammoCollision(Weapon *weapon, Vector3Dg position, Vector3Dg direction)
{
	//calculate the end point of the ammo's way
	Vector3Dg end = position + (direction * weapon->getRange());
	//random shift in direction
	float shift = 25.0f - (rand() / (double)RAND_MAX) * 50.0f;
	end.x += direction.z * shift;
	shift = 25.0f - (rand() / (double)RAND_MAX) * 50.0f;
	end.z += direction.x * shift;
	//calculate the direction once again 'cause we gonna need it
	direction = end - position;
	direction.normalize();
	//calculate the boundaries of potentially affected area
	int8_t beginx = (int8_t)(position.x / 100), beginz = (int8_t)(-position.z / 100);
	int8_t endx = (int8_t)(end.x / 100), endz = (int8_t)(-end.z / 100);
	//make sure that begining values are lesser than ending ones
	if(endx < 0)
		endx = 0;
	if(endz < 0)
		endz = 0;
	if(beginx > endx) {
		int8_t tmp = beginx;
		beginx = endx;
		endx = tmp;
	}
	if(beginz > endz) {
		int8_t tmp = beginz;
		beginz = endz;
		endz = tmp;
	}
	//make sure we are not out of bounds
	if(beginx < 0)
		beginx = 0;
	if(beginz < 0)
		beginz = 0;
	if(endx > 63)
		endx = 63;
	if(endz > 63)
		endz = 63;

	//which walls to check
	uint32_t count = 0, check[2];
	//north
	if(direction.z < 0.0f) {
		check[count] = 0;
		count++;
	}
	//south
	else if(direction.z > 0.0f) {
		check[count] = 2;
		count++;
	}
	//east
	if(direction.x > 0.0f) {
		check[count] = 1;
		count++;
	}
	//west
	else if(direction.x < 0.0f) {
		check[count] = 3;
		count++;
	}

	//distaince of the closest object
	float distaince = weapon->getRange();
	GameObject *closestobject = 0;
	uint32_t closestobjoffset;

	//check all fields for collisions
	for(int8_t c = beginx; c <= endx; c++) {
		for(int8_t r = beginz; r <= endz; r++) {
			//position's offset in the map buffer
			uint32_t offset = c * 8 + r * 512;

			//get info about existing walls in the square
			uint16_t walls = mLevelMap[offset + 4];
			for(uint32_t i = 0; i < count; ++i) {
				if(not (walls & (1 << check[i])))
					continue;
				Mesh3DgUID meshnum = mLevelMap[offset + check[i]];
				GameObject *obj = mObjects->getGameObject(meshnum);
				if(obj->getAmmoCollisionType() == ACTYPE_NONE)
					continue;
				float dist = 0.0f;
				switch(obj->getType()) {
					case GOTYPE_WALL:
						dist = wallCollisionAmmo(position, end, c, r, check[i], obj);
						break;
					case GOTYPE_DOOR: {
						//only if not open
						if(isDoorOpen((GODoor *)obj))
							dist = wallCollisionAmmo(position, end, c, r, check[i], obj);
						break;
					}
					default:
						break;
				}
				if(dist != 0.0f and dist < distaince) {
					distaince = dist;
					closestobject = obj;
					closestobjoffset = offset;
				}
			}

			//sprites
			Mesh3DgUID meshnum = mLevelMap[offset + 5];
			if(mLevelMap[offset + 4] & 16) {
				GameObject *obj = mObjects->getGameObject(meshnum);
				if(obj->getAmmoCollisionType() != ACTYPE_NONE) {
					float dist = spriteCollisionAmmo(position, end, c, r, weapon->getRange());
					if(dist != 0.0f and dist < distaince) {
						distaince = dist;
						closestobject = obj;
						closestobjoffset = offset;
					}
				}
			}
		}
	}

	//now check for enemies
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++) {
		GameObject *obj = (GameObject *)*i;
		if(obj->getAmmoCollisionType() != ACTYPE_NONE) {
			Vector3Dg enemypos = obj->getMesh()->geomCenter();
			enemypos.y = 0.0f;
			float dist = enemyCollisionAmmo(position, end, enemypos, weapon->getRange());
			if(dist != 0.0f and dist < distaince) {
				distaince = dist;
				closestobject = obj;
			}
		}
	}

	//check the player
	float playerDistaince = playerCollisionAmmo(position, end, weapon->getRange());
	//if the distaince to the player is less than the distaince to the closest object
	//then the player is hit
	if(playerDistaince > 0.0 and playerDistaince < distaince) {
		playerHit(weapon->getPower(), direction);
		return;
	}

	//if any object has been hit
	if(closestobject) {
		Vector3Dg explosionpos = position + (direction * (distaince - 5.0f));
		switch(closestobject->getAmmoCollisionType()) {
			case ACTYPE_UNBREAKABLE:
				displayExplosion(explosionpos, EXPLT_SMALL, 0, true);
				return;
			case ACTYPE_EXPLODABLE: {
				if(closestobject->getType() != GOTYPE_SPRITE)
					return;
				//if run out of the hp
				if(((GOSprite *)closestobject)->decreaseHP(weapon->getPower()) <= 0) {
					closestobject->getMesh()->setProperty(MRP_RENDERABLE, false);
					//rem info in map
					mLevelMap[closestobjoffset + 4] &= ~16;
					//if there is an equipment near the explodable object then destroy it to
					if(mLevelMap[closestobjoffset + 4] & 32) {
						Mesh3DgUID meshnum = mLevelMap[closestobjoffset + 6];
						mObjects->mMeshes[meshnum]->setProperty(MRP_RENDERABLE, false);
						//rem info in map
						mLevelMap[closestobjoffset + 4] &= ~32;
					}
					explosionCollision(explosionpos);
					displayHugeExplosion(explosionpos, direction);
					mSDL->playSound(mCollisionSounds[CSND_EXPLOSION], 3, calculateVolume(explosionpos, VOLDIST_TYPE1));
				}
				//else - only display the ammo's explosion
				else displayExplosion(explosionpos, EXPLT_SMALL, 0, true);
				return;
			}
			case ACTYPE_ENEMY: {
				displayExplosion(explosionpos, EXPLT_SMALL, 0, true);

				GOEnemy *enemy = (GOEnemy *)closestobject;
				//decrease the HP and die, if necessary
				//(yes, in the original game, the enemy died after his HP went below 0)
				if(enemy->isBurning() or not enemy->isAlive())
					return;
				if(enemy->hit(weapon->getPower(), mCameraPosition) < 0) {
					enemy->die();
					//sometimes leave a present
					leaveAmmo(enemy);
					//blood
					spreadBlood(enemy, direction, distaince);
					mSDL->playSound(mEnemySounds[ESND_DEATH], 4, calculateVolume(explosionpos, VOLDIST_TYPE3));
				} else hitRoar(enemy);
				checkEnemyCollisions(enemy);
				enemy->move();
				return;
			}
			default:
				return;
		}
	}
}

void IGameCtrl::displayHugeExplosion(Vector3Dg &position, Vector3Dg &direction)
{
	displayExplosion(position, EXPLT_BIG, 0, false);

	Vector3Dg vec(-direction.z, direction.y, direction.x);
	Vector3Dg vec2 = position - (direction * 10.0f);
	displayExplosion(vec2, EXPLT_BIG, 200, true);
	vec2 = position - (vec * 100.0f) - (direction * 10.0f);
	displayExplosion(vec2, EXPLT_BIG, 200, true);
	vec2 = position + (vec * 100.0f) - (direction * 10.0f);
	displayExplosion(vec2, EXPLT_BIG, 200, true);
	vec2 = position - (vec * 80.0f) + (direction * 10.0f);
	displayExplosion(vec2, EXPLT_BIG, 200, true);
	vec2 = position + (vec * 80.0f) + (direction * 10.0f);
	displayExplosion(vec2, EXPLT_BIG, 200, true);
}

void IGameCtrl::displayExplosion(Vector3Dg &position, ExplosionType type, uint32_t delay,
								bool yshift)
{
	Mesh3Dg *mesh = mObjects->createMesh(mObjects->generateMeshUID());
	if(not mesh)
		return;
	uint32_t time = 0;
	switch(type) {
		case EXPLT_SMALL:
		case EXPLT_ELECTRIC:
			time = 600;
			break;
		case EXPLT_BIG:
			time = 800;
			break;
		break;
	}
	TemporaryObject *obj = mObjects->createTempGameObject(mesh, time);
	if(not obj) {
		mObjects->destroyMesh(mesh->getUID());
		return;
	}
	//set the animation start delay
	if(delay) {
		obj->setAnimStartDelay(delay);
		mesh->setProperty(MRP_RENDERABLE, false);
	}
	//render properties for mesh
	mesh->setProperty(MRP_TRANSPARENT, true);
	mesh->setProperty(MRP_Z_MINUS_OFFSET, true);
	//position
	float y = position.y;
	if(yshift)
		y += 5.0f - (rand() / (double)RAND_MAX) * 10.0f;
	mesh->setPosition(position.x, y, position.z);
	switch(type) {
		case EXPLT_SMALL: {
			//mesh base
			mesh->setBase(mSmallExplosionBase);
			//animation
			mesh->mAnimation.dwTexAnimTime = time;
			mesh->mAnimation.dwBeginMtl = 85;
			mesh->mAnimation.dwNumMtls = 3;
			break;
		}
		case EXPLT_BIG: {
			//mesh base
			mesh->setBase(mBigExplosionBase);
			//animation
			mesh->mAnimation.dwTexAnimTime = time;
			mesh->mAnimation.dwBeginMtl = 81;
			mesh->mAnimation.dwNumMtls = 4;
			break;
		}
		case EXPLT_ELECTRIC: {
			//mesh base
			mesh->setBase(mSmallExplosionBase);
			//animation
			mesh->mAnimation.dwTexAnimTime = time;
			mesh->mAnimation.dwBeginMtl = 88;
			mesh->mAnimation.dwNumMtls = 3;
			break;
		}
	}

	//add the object to the render and temp-objects lists
	mTempObjects.push_back(obj);
	mOGL->addMeshToRenderList(mesh);
}

bool IGameCtrl::cardCollision(UsableItemType eCardColor)
{
	uint32_t dwPosOffset = mPosition[0] * 8 + mPosition[1] * 512;
	uint32_t dwFaceDir = getFaceDir();
	Mesh3DgUID dwMeshNum = mLevelMap[dwPosOffset + dwFaceDir];
	//if there is no wall
	if(not (mLevelMap[dwPosOffset + 4] & (1 << dwFaceDir))) {
		mOGL->mOSD.addText(mTexts[NO_SLOT_TEXTS].getText());
		return false;
	}
	//check for slot
	GameObject *obj = mObjects->getGameObject(dwMeshNum);
	if(obj->getType() == GOTYPE_DOOR) {
		mOGL->mOSD.addText(mTexts[NO_SLOT_TEXTS].getText());
		return false;
	}

	assert(obj->getType() == GOTYPE_WALL);

	Mesh3Dg *pSlotMesh = ((GOWall *)obj)->getSlotMesh();
	if(not pSlotMesh) {
		mOGL->mOSD.addText(mTexts[NO_SLOT_TEXTS].getText());
		return false;
	}
	//check if the slot isn't used
	int32_t iMtlNum = pSlotMesh->mMaterialNumber;
	switch(iMtlNum) {
		case 51:
		case 142:
		case 145:
		case 148:
			break;
		case 52:
		case 143:
		case 146:
		case 149:
			mOGL->mOSD.addText(mTexts[SLOT_USED_TEXTS].getText());
			return false;
		default:
			mOGL->mOSD.addText(mTexts[NO_SLOT_TEXTS].getText());
			return false;
	}
	//check for action desc
	MapDataPtr pData = &mLevelMap[0x8000];
	uint32_t dwDataOffset = getDataOffset(((GOWall *)obj)->getSlotActionDataOffset() + 1, pData);

	//if no action desc found then the slot is broken
	if(dwDataOffset == 0) {
		mOGL->mOSD.addText(mTexts[SLOT_BROKEN_TEXTS].getText());
		return false;
	}
	dwDataOffset++;
	//if wrong color
	uint16_t wCardColor;
	switch(eCardColor) {
		case UIT_RED_CARD:
			wCardColor = 1;
		break;
		case UIT_GREEN_CARD:
			wCardColor = 2;
		break;
		case UIT_BLUE_CARD:
			wCardColor = 3;
		break;
		//should not happen
		default:
			return false;
	}
	if(pData[dwDataOffset] != wCardColor) {
		mOGL->mOSD.addText(mTexts[CARD_WRONG_TEXTS].getText());
		return false;
	}
	mOGL->mOSD.addText(mTexts[CARD_INSERT_TEXTS].getText());
	//change material
	pSlotMesh->mMaterialNumber++;
	((GOWall *)obj)->setBasicSlotMaterial(((GOWall *)obj)->getBasicSlotMaterial() + 1);
		
	dwDataOffset++;PRE_LOOP();
	//do action
	while(pData[dwDataOffset] != 0xFFFF) {
		doActions(dwDataOffset, pData);
		dwDataOffset += 2;
	}AFTER_LOOP();
	return true;
}

void IGameCtrl::doActions(uint32_t dwDataOffset, MapDataPtr pData)
{
	uint8_t btByte = (uint8_t)pData[dwDataOffset];
	uint8_t btZ = (uint8_t)(pData[dwDataOffset + 1] / 512);
	uint8_t btX = (uint8_t)((pData[dwDataOffset + 1] % 512) / 8);
	switch(btByte) {
		//open the door
		case 0: {
			Mesh3DgUID dwMeshNum = mLevelMap[pData[dwDataOffset + 1]];
			//if closed then open
			GODoor *door = (GODoor *)mObjects->getGameObject(dwMeshNum);
			uint32_t uiDoorType = door->getDoorType();
			animDoorOpen(door, uiDoorType);
			//if double door - open the second part too
			if(uiDoorType == DOORTYPE_HORIZONTAL) {
				door = (GODoor *)mObjects->getGameObject(dwMeshNum+1);
				animDoorOpen(door, DOORTYPE_UNDEFINED);
			}
			break;
		}
		//close the door
		case 1: {
			Mesh3DgUID dwMeshNum = mLevelMap[pData[dwDataOffset + 1]];
			//if opened then close
			GODoor *door = (GODoor *)mObjects->getGameObject(dwMeshNum);
			uint32_t uiDoorType = door->getDoorType();
			animDoorClose(door, uiDoorType);
			//if double door
			if(uiDoorType == DOORTYPE_HORIZONTAL) {
				door = (GODoor *)mObjects->getGameObject(dwMeshNum+1);
				animDoorClose(door, DOORTYPE_UNDEFINED);
			}
			break;
		}
		//set door blocade
		case 2:
			mLevelMap[btZ * 512 + btX * 8 + 4] |= 64;
			break;
		//remove door blocade
		case 3:
			mLevelMap[btZ * 512 + btX * 8 + 4] &= ~64;
			break;
		//set/remove wall
		case 4: {
			uint8_t btWallMtl = (uint8_t)(pData[dwDataOffset] >> 8);
			MapDataType wOffset = pData[dwDataOffset + 1];
			Mesh3DgUID dwMeshNum = mLevelMap[wOffset];
			uint8_t btZ = (uint8_t)(wOffset / 512);
			uint8_t btX = (uint8_t)((wOffset % 512) / 8);
			//calculate the offset once again
			uint16_t wOffset2 = (uint16_t)(btX * 8 + btZ * 512);
			uint16_t wWallNum = (uint16_t)(wOffset - wOffset2);
			//if the wall is visible
			bool bIsWall = (mLevelMap[wOffset2 + 4] & (1 << wWallNum)) ? true : false;
			//remove the wall only if it is there
			if(not btWallMtl and bIsWall) {
				//don't render
				mObjects->mMeshes[dwMeshNum]->setProperty(MRP_RENDERABLE, false);
				//remove info in the map
				mLevelMap[wOffset2 + 4] &= ~(1 << wWallNum);
				break;
			}
			//if only changing the wall then just switch materials
			if(btWallMtl and bIsWall) {
				Mesh3Dg *mesh = mObjects->mMeshes[dwMeshNum];
				mesh->mMaterialNumber = btWallMtl - 1;
				GameObject *obj = (GameObject *)mesh->getChild();
				//collision type for new wall based on the material
				if((btWallMtl - 1) == 26) {
					obj->setPlayerCollisionType(PCTYPE_NONE);
					obj->setAmmoCollisionType(ACTYPE_NONE);
				} else {
					obj->setPlayerCollisionType(PCTYPE_BUMP);
					obj->setAmmoCollisionType(ACTYPE_UNBREAKABLE);
				}
				break;
			}
			//set the wall, only if it isn't there
			if(btWallMtl and not bIsWall) {
				//make renderable
				mObjects->mMeshes[dwMeshNum]->setProperty(MRP_RENDERABLE, true);
				//add info in map
				mLevelMap[wOffset2 + 4] |= (1 << wWallNum);
				break;
			}
			break;
		}
		//set/remove equipment
		case 5: {
			uint8_t btEqMtl = (uint8_t)(pData[dwDataOffset] >> 8);
			MapDataType wOffset = pData[dwDataOffset + 1];
			//calculate position offset w/o shift
			//(in action data the offset is sometimes shifted (why?))
			uint8_t btZ = (uint8_t)(wOffset / 512);
			uint8_t btX = (uint8_t)((wOffset % 512) / 8);
			//calculate offset shifted by 6
			wOffset = (uint16_t)(btX * 8 + btZ * 512 + 6);
			Mesh3DgUID dwMeshNum = mLevelMap[wOffset];
			bool bIsEq = (mLevelMap[wOffset - 2] & 32) ? true : false;
			//remove equipment - only if it is there
			if(not btEqMtl and bIsEq) {
				mObjects->mMeshes[dwMeshNum]->setProperty(MRP_RENDERABLE, false);
				//rem info in map
				mLevelMap[wOffset - 2] &= ~32;
				break;
			}
			//change equipment
			if(btEqMtl and bIsEq) {
				mObjects->mMeshes[dwMeshNum]->mMaterialNumber = btEqMtl - 1;
				break;
			}
			//set equipment
			if(btEqMtl and not bIsEq) {
				//make renderable
				mObjects->mMeshes[dwMeshNum]->setProperty(MRP_RENDERABLE, true);
				//add info in map
				mLevelMap[wOffset - 2] |= 32;
				break;
			}
			break;
		}
		//set/remove sprite
		case 6: {
			uint8_t btSpriteMtl = (uint8_t)(pData[dwDataOffset] >> 8);
			MapDataType wOffset = pData[dwDataOffset + 1];
			//calculate position offset w/o shift
			//(in action data the offset is sometimes shifted (why?))
			uint8_t btZ = (uint8_t)(wOffset / 512);
			uint8_t btX = (uint8_t)((wOffset % 512) / 8);
			//calculate offset shifted by 5
			wOffset = (uint16_t)(btX * 8 + btZ * 512 + 5);
			Mesh3DgUID dwMeshNum = mLevelMap[wOffset];
			bool bIsSprite = (mLevelMap[wOffset - 1] & 16) ? true : false;
			//remove sprite - only if it is there
			if(not btSpriteMtl and bIsSprite) {
				mObjects->mMeshes[dwMeshNum]->setProperty(MRP_RENDERABLE, false);
				//rem info in map
				mLevelMap[wOffset - 1] &= ~16;
				break;
			}
			//change sprite
			if(btSpriteMtl and bIsSprite) {
				--btSpriteMtl;
				Mesh3Dg *mesh = mObjects->mMeshes[dwMeshNum];
				mesh->mMaterialNumber = btSpriteMtl;
				GameObject *obj = (GameObject *)mesh->getChild();
				switch(btSpriteMtl) {
					case 30:
						obj->setPlayerCollisionType(PCTYPE_TELEPORT);
						break;
					case 27:
					case 28:
						obj->setPlayerCollisionType(PCTYPE_STOP);
						obj->setAmmoCollisionType(ACTYPE_UNBREAKABLE);
						break;
					case 44:
						obj->setPlayerCollisionType(PCTYPE_STOP);
						obj->setAmmoCollisionType(ACTYPE_EXPLODABLE);
						break;
					default:
						obj->setPlayerCollisionType(PCTYPE_NONE);
						break;
				}
				break;
			}
			//set sprite
			if(btSpriteMtl and not bIsSprite) {
				//make renderable
				mObjects->mMeshes[dwMeshNum]->setProperty(MRP_RENDERABLE, true);
				//add info in map
				mLevelMap[wOffset - 1] |= 16;
				break;
			}
			break;
		}
	}
}

void IGameCtrl::enemyBackWallCollision(GOEnemy *enemy, Mesh3Dg *wallMesh, int8_t wallDir, float ray)
{
	//calculate normal vector for the wall
	Vector3Dg normal(0.0f, 0.0f, 0.0f);
	switch(wallDir) {
		case 0: normal.z = -1.0f; break;
		case 1: normal.x = 1.0f; break;
		case 2: normal.z = 1.0f; break;
		case 3: normal.x = -1.0f; break;
	}
	//coordinates of the first vertex in the first triangle (a pointer is enough)
	const Point3Dg *pt = wallMesh->getBase()->getVCoordsStream();
	//the distaince between the enemy and the wall
	float dist = -normal.x * (pt->x - enemy->getMesh()->geomCenter().x)
					-normal.z * (pt->z - enemy->getMesh()->geomCenter().z);
	//if distance from wall is lesser than the enemy's ray
	if(dist < ray) {
		Vector3Dg vec = enemy->getActualVelocity();
		vec.normalize();
		//if angle between direction and wall normal is < 45 deg
		//(can happen after going through a wall from the back)
		if(vec.dotProduct(normal) > 0.0f)
			return;
		//slide on the wall
		vec.x = vec.y = vec.z = 0.0f;
		enemy->setActualVelocity(vec);
	}
}

void IGameCtrl::enemyWallCollision(GOEnemy *enemy, Mesh3Dg *wallMesh, int8_t wallDir, float ray)
{
	//calculate normal vector for the wall
	Vector3Dg normal(0.0f, 0.0f, 0.0f);
	switch(wallDir) {
		case 0: normal.z = 1.0f; break;
		case 1: normal.x = -1.0f; break;
		case 2: normal.z = -1.0f; break;
		case 3: normal.x = 1.0f; break;
	}
	//coordinates of the first vertex in the first triangle (a pointer is enough)
	const Point3Dg *pt = wallMesh->getBase()->getVCoordsStream();
	//the distaince between the enemy and the wall
	float dist = -normal.x * (pt->x - enemy->getMesh()->geomCenter().x)
					-normal.z * (pt->z - enemy->getMesh()->geomCenter().z);
	//if distance from wall is lesser than the enemy's ray
	if(dist < ray) {
		Vector3Dg vec = enemy->getActualVelocity();
		vec.normalize();
		//if angle between direction and wall normal is < 45 deg
		//(can happen after going through a wall from the back)
		if(vec.dotProduct(normal) > SQR2BY2)
			return;
		//slide on the wall
		normal *= ray - dist;
		vec = enemy->getActualVelocity() + normal;
		enemy->setActualVelocity(vec);
	}
}

bool IGameCtrl::bulletEdgesCollision(bool *edges, Vector3Dg &begin, Vector3Dg &end, int8_t x, int8_t z, float edgeRay, Vector3Dg &destination)
{
	Vector3Dg way = end - begin;

	//calculate coordinates of all edges and put them into arrays
	float posX = x * 100.0f;
	float posZ = -(z * 100.0f);
	float edgesXCoords[] = {posX, posX + 100.0f, posX + 100.0f, posX};
	float edgesZCoords[] = {posZ - 100.0f, posZ - 100.0f, posZ, posZ};
	//check collisions with all edges
	for(int i = 0; i < 4; i++) {
		if(not edges[i])
			continue;
		//a vector connecting the begin point with the edge
		Vector3Dg vec(edgesXCoords[i] - begin.x, 0.0f, edgesZCoords[i] - begin.z);
		//calculate a projection of the edge on the way vector in coords relative to the
		//vector {begin -> 0.0f; end -> 1.0f)
		float projectionPos = way.dotProduct(vec) / way.lengthPow2();
		//calculate the coords of a point that we gonna use to calculate the distaince
		//between the way and the edge (store it in the destination vector);
		//the distaince will be calculated either from the begining of the way,
		if(projectionPos <= 0.0f)
			destination = begin;
		//the ending of the way,
		else if(projectionPos >= 1.0f)
			destination = end;
		//or from a point being the edge projected on the way
		else {
			//calculate the coords of the projection:
			//shorten the way vector so that it points to the projection position
			way *= projectionPos;
			//the projected point itself:
			destination = begin + way;
		}
		//finally, calculate the distaince between the point and the edge
		//(we can use the power of 2 - less calculations (no sqrtf))
		float distaincePow2 = (edgesXCoords[i] - destination.x) * (edgesXCoords[i] - destination.x) +
							(edgesZCoords[i] - destination.z) * (edgesZCoords[i] - destination.z);
		if(distaincePow2 < edgeRay * edgeRay)
			return true;
	}
	return false;
}

void IGameCtrl::enemyEdgesCollision(bool *edges, GOEnemy *enemy, int8_t x, int8_t z, float ray)
{
	float posX = x * 100.0f - enemy->getMesh()->geomCenter().x;
	float posZ = -(z * 100.0f) - enemy->getMesh()->geomCenter().z;

	if(edges[0]) {
		float x = posX;
		float z = posZ - 100.0f;
		if(x * x + z * z < ray * ray) {
			float length = sqrtf(x * x + z * z);
			length = ray / length - 1.0f;
			Vector3Dg vec = enemy->getActualVelocity();
			vec.x -= x * length;
			vec.z -= z * length;
			enemy->setActualVelocity(vec);
			return;
		}
	}
	if(edges[1]) {
		float x = posX + 100.0f;
		float z = posZ - 100.0f;
		if(x * x + z * z < ray * ray) {
			float length = sqrtf(x * x + z * z);
			length = ray / length - 1.0f;
			Vector3Dg vec = enemy->getActualVelocity();
			vec.x -= x * length;
			vec.z -= z * length;
			enemy->setActualVelocity(vec);
			return;
		}
	}
	if(edges[2]) {
		float x = posX + 100.0f;
		float z = posZ;
		if(x * x + z * z < ray * ray) {
			float length = sqrtf(x * x + z * z);
			length = ray / length - 1.0f;
			Vector3Dg vec = enemy->getActualVelocity();
			vec.x -= x * length;
			vec.z -= z * length;
			enemy->setActualVelocity(vec);
			return;
		}
	}
	if(edges[3]) {
		float x = posX;
		float z = posZ;
		if(x * x + z * z < ray * ray) {
			float length = sqrtf(x * x + z * z);
			length = ray / length - 1.0f;
			Vector3Dg vec = enemy->getActualVelocity();
			vec.x -= x * length;
			vec.z -= z * length;
			enemy->setActualVelocity(vec);
		}
	}
}

uint16_t IGameCtrl::wallCollision(GameObject *obj, int8_t wallDir)
{
	//calculate normal vector for the wall
	Vector3Dg normal(0.0f, 0.0f, 0.0f);
	switch(wallDir) {
		case 0: normal.z = 1.0f; break;
		case 1: normal.x = -1.0f; break;
		case 2: normal.z = -1.0f; break;
		case 3: normal.x = 1.0f; break;
	}
	Mesh3Dg *mesh = obj->getMesh();
	//coordinates of the first vertex in the first triangle (a pointer is enough)
	const Point3Dg *pt = mesh->getBase()->getVCoordsStream();

	float fDist = -normal.x * (pt->x - mPrevCamPosition.x - mVelocity.x)
					-normal.z * (pt->z - mPrevCamPosition.z - mVelocity.z);
	//if distance from wall is lesser than camera ray (fDist is always > 0)
	if(fDist < mCamRay) {
		Vector3Dg sNormalizedV = mVelocity;
		sNormalizedV.normalize();
		float fDotP = sNormalizedV.dotProduct(normal);
		//if angle between velocity and wall normal is < 45 deg
		//(can happen after going through a wall from the back)
		if(fDotP > SQR2BY2)
			return 0;
		//bump the wall
		if(fDotP < -SQR2BY2) {
			intfloat ix, iz;
			//module
			ix.f = mVelocity.x; ix.i &= 0x7FFFFFFF;
			iz.f = mVelocity.z; iz.i &= 0x7FFFFFFF;
			//calculate the auxiliary velocity that will be used to
			//simulate the inertia after bumping the wall
			//(we don't zero the regular velocity here 'cause we want the
			//bump to begin at the position to which the player actually
			//got; the regular velocity will be zeroed as soon as the
			//inertia simulation starts)
			mAuxiliaryV.x = mVelocity.x + normal.x * ix.f * 2;
			mAuxiliaryV.z = mVelocity.z + normal.z * iz.f * 2;
			mAuxVTime = 250;
			int32_t iMtlNum = mesh->mMaterialNumber;
			//if door collision
			if(obj->getType() == GOTYPE_DOOR) {
				//and its material number
				iMtlNum = ((GODoor *)obj)->getFrameMesh()->mMaterialNumber;
				switch(iMtlNum) {
					//closed or opening?
					case 14:
					case 17:
						mOGL->mOSD.addText(mTexts[BUMP_DOOR_TEXTS].getText());
						break;
					case 16:
					case 19:
						mOGL->mOSD.addText(mTexts[BUMP_OPENING_TEXTS].getText());
						break;
				}
			}
			//normal wall collision
			else mOGL->mOSD.addText(mTexts[BUMP_WALL_TEXTS].getText());
			excite();
			mSDL->playSound(mCollisionSounds[CSND_WALL_BUMP], 1);
			if(mVisionNoise)
				displayNoise();
			if(mHighDificulty and not mImmortal)
				return 1;
			return 0;
		}
		//if no bumping then just move the player away from the wall to a
		//distaince equal to the camera's ray
		normal *= mCamRay - fDist;
		mVelocity += normal;
	}
	return 0;
}

void IGameCtrl::edgesCollision(bool *pEdges)
{
	float fPosX = mPosition[0] * 100.0f - mPrevCamPosition.x - mVelocity.x;
	float fPosZ = -(mPosition[1] * 100.0f) - mPrevCamPosition.z - mVelocity.z;

	if(pEdges[0]) {
		float x = fPosX;
		float z = fPosZ - 100.0f;
		if(x * x + z * z < mCamRay * mCamRay) {
			float length = sqrtf(x * x + z * z);
			length = mCamRay / length - 1.0f;
			mVelocity.x -= x * length;
			mVelocity.z -= z * length;
			return;
		}
	}
	if(pEdges[1]) {
		float x = fPosX + 100.0f;
		float z = fPosZ - 100.0f;
		if(x * x + z * z < mCamRay * mCamRay) {
			float length = sqrtf(x * x + z * z);
			length = mCamRay / length - 1.0f;
			mVelocity.x -= x * length;
			mVelocity.z -= z * length;
			return;
		}
	}
	if(pEdges[2]) {
		float x = fPosX + 100.0f;
		float z = fPosZ;
		if(x * x + z * z < mCamRay * mCamRay) {
			float length = sqrtf(x * x + z * z);
			length = mCamRay / length - 1.0f;
			mVelocity.x -= x * length;
			mVelocity.z -= z * length;
			return;
		}
	}
	if(pEdges[3]) {
		float x = fPosX;
		float z = fPosZ;
		if(x * x + z * z < mCamRay * mCamRay) {
			float length = sqrtf(x * x + z * z);
			length = mCamRay / length - 1.0f;
			mVelocity.x -= x * length;
			mVelocity.z -= z * length;
		}
	}
}

void IGameCtrl::openDoor(Mesh3DgUID dwMeshNum, GODoor *door)
{
	//door material number
	uint32_t uiDoorType = door->getDoorType();
	//open the door only if not fully opened
	if(door->getMesh()->mMaterialNumber + 1 != door->getFrameMesh()->mMaterialNumber) {
		animDoorOpen(door, uiDoorType);
		//if double door
		if(uiDoorType == DOORTYPE_HORIZONTAL) {
			door = (GODoor *)mObjects->getGameObject(dwMeshNum+1);
			animDoorOpen(door, DOORTYPE_UNDEFINED);
		}
	}
}

void IGameCtrl::closePrevDoor(uint16_t wPrevOffset, uint16_t wCurrentPosOffset)
{
	bool bCurrentPosLocked = mLevelMap[wCurrentPosOffset + 4] & 64;
	//check all doors at previous position
	for(uint32_t i = 0; i < 4; i++) {
		//i-th mesh on the previous position
		Mesh3DgUID dwMeshNum = mLevelMap[wPrevOffset + i];
		if(not dwMeshNum)
			continue;
		//number of a mesh on the current position and of direction oposite
		//to the dwMeshNum (ie. if dwMeshNum is faced EAST then dwMeshNum2 is
		//faced WEST)
		Mesh3DgUID dwMeshNum2 = mLevelMap[wCurrentPosOffset + ((i+2) % 4)];
		//if it is the same mesh then these are the door the player just have
		//went through; in such case don't close this door if they aren't locked
		//on the current position
		if((dwMeshNum == dwMeshNum2) and not bCurrentPosLocked)
			continue;
		//if there are any
		GameObject *obj = mObjects->getGameObject(dwMeshNum);
		if(obj->getType() == GOTYPE_DOOR) {
			GODoor *door = (GODoor *)obj;
			uint32_t uiDoorType = door->getDoorType();
			animDoorClose(door, uiDoorType);
			//if double door
			if(uiDoorType == DOORTYPE_HORIZONTAL) {
				door = (GODoor *)mObjects->getGameObject(dwMeshNum+1);
				animDoorClose(door, DOORTYPE_UNDEFINED);
			}
		}
	}
}

void IGameCtrl::animDoorOpen(GODoor *door, uint32_t uiDoorType)
{
	//don't start the 'opening' animation if it is already in progress
	Mesh3Dg *mesh = door->getMesh();
	if(mesh->mAnimation.iAnimStatus != ANIM_FORWARD and
				mesh->mAnimation.iElapsedTime != (int32_t)mesh->mAnimation.dwGeomAnimTime) {
		mSDL->playSound(mDoorSounds[uiDoorType], 6, calculateVolume(mesh->geomCenter(), VOLDIST_TYPE1));
		mesh->mAnimation.iAnimStatus = ANIM_FORWARD;
		door->getFrameMesh()->mMaterialNumber = mesh->mMaterialNumber + 2;
		door->getFrameMesh()->setProperty(MRP_RENDERABLE, true);
	}
}

void IGameCtrl::animDoorClose(GODoor *door, uint32_t uiDoorType)
{
	Mesh3Dg *mesh = door->getMesh();
	if(mesh->mAnimation.iAnimStatus != ANIM_BACKWARD and mesh->mAnimation.iElapsedTime != 0) {
		mSDL->playSound(mDoorSounds[uiDoorType], 6, calculateVolume(mesh->geomCenter(), VOLDIST_TYPE1));
		mesh->mAnimation.iAnimStatus = ANIM_BACKWARD;
		door->getFrameMesh()->mMaterialNumber = mesh->mMaterialNumber + 2;
	}
}

bool IGameCtrl::enemyObjectCollision(Vector3Dg &objPos, GOEnemy *enemy, float objRay, float ray)
{
	Vector3Dg enemyVelocity = enemy->getActualVelocity();
	//a vector connecting the enemy and the object
	Vector3Dg vec = objPos - enemy->getMesh()->geomCenter() - enemyVelocity;
	//it's length
	float dist = vec.length();
	//if distaince is greater than the enemy's ray - there is no collision
	if(dist < objRay + ray) {
		//compare the enemy's direction with the calculated vector's direction
		vec.normalize();
		float dotProduct = enemy->getDirection().dotProduct(vec);
		//if the angle between the vectors is less than 90 deg
		if(dotProduct > 0.0f) {
			vec.x = vec.y = vec.z = 0.0f;
			enemy->setActualVelocity(vec);
			return true;
		}
	}
	return false;
}

void IGameCtrl::enemyCollision(GOEnemy *enemyToCheck)
{
	//check all enemies
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++) {
		GOEnemy *enemy = *i;
		//don't check against itself
		if(enemy == enemyToCheck or not enemy->isAlive())
			continue;
		Vector3Dg pos = enemy->getMesh()->geomCenter();
		enemyObjectCollision(pos, enemyToCheck, 25.0f, 25.0f);
	}
}

bool IGameCtrl::spriteCollision(Mesh3Dg *mesh)
{
	//a vector connecting the player and the mesh
	Vector3Dg vec(mesh->mTranslation.R4C1, 0.0f, mesh->mTranslation.R4C3);
	vec = vec - mPrevCamPosition - mVelocity;
	//it's lenght
	float dist = vec.length();
	//if distaince is greater than camera ray - there is no collision
	if(dist < mCamRay) {
		float length = mVelocity.length();
		if(length == 0.0f)
			return true;
		mVelocity *= (length - mCamRay + dist) / length;
		return true;
	}
	return false;
}

uint16_t IGameCtrl::enemyCollision()
{
	//check all enemies
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++) {
		GOEnemy *enemy = *i;
		if(enemy->getPlayerCollisionType() != PCTYPE_NONE and spriteCollision(enemy->getMesh())) {
			if(enemy->isBurning() and mPlayerBurnDiff >= PLAYER_BURN_DELAY) {
				mPlayerBurnDiff = 0;
				mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_HOT));
				mSDL->playSound(mCollisionSounds[CSND_TOUCH_FLAMES], 1);
				excite();
				if(mHighDificulty and not mImmortal)
					return 1;
			}
			return 0;
		}
	}
	return 0;
}

void IGameCtrl::explosionCollision(Vector3Dg &position)
{
	//if to close to the explosion then decrease energy
	Vector3Dg vec = mCameraPosition - position;
	float dist = vec.length();
	vec /= dist;

	if(dist < mCamRay+50.0f) {
		if(dist < mCamRay+30.0f) {
			playerHit(25);
			mAuxiliaryV = vec * 2.0f;
		} else {
			playerHit(5);
			mAuxiliaryV = vec;
		}
		//calculate the auxiliary velocity that will be used to
		//simulate the inertia effect
		mAuxiliaryV.y = 0.0f;
		mAuxVTime = 250;
	}

	//on which field the explosion took place?
	uint32_t x = ((int)position.x/100);
	uint32_t z = -((int)position.z/100);
	uint32_t posOffset = x * 8 + z * 512;

	//burn enemies that stand to close
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++) {
		GOEnemy *enemy = *i;
		if(not enemy->isAlive())
			continue;
		//get the enemy position
		Vector3Dg enemypos = enemy->getMesh()->geomCenter();
		enemypos.y = 0.0f;
		//offset of the field on which the enemy is standing
		uint32_t ex = ((int)enemypos.x/100);
		uint32_t ez = -((int)enemypos.z/100);
		uint32_t enemyPosOffset = ex * 8 + ez * 512;
		//if the enemy is standing on other field than the explosion took place, then check
		//if there is a wall between the explosion and the enemy
		if(enemyPosOffset != posOffset) {
			MapDataType walls = mLevelMap[posOffset + 4];
			if((walls & 1 and ez > z) or (walls & 2 and ex > x) or
				(walls & 4 and ez < z) or (walls & 8 and ex < x))
				continue;
		}

		//calculate a vector from the explosion's position to the enemy
		vec = enemypos - position;
		dist = vec.length();
		vec /= dist;
		//burn if to close
		if(dist < 75.0f) {
			if(enemy->burn()) {
				//turn off other fire sounds
				unregisterFireSounds();
				mSDL->playSound(mEnemySounds[ESND_FIRE], 5, 128, -1);
				enemy->registerSound(5);
			}
		}
	}
}

void IGameCtrl::playerHit(uint16_t lostHP, Vector3Dg direction)
{
	//lose some energy
	if(not mImmortal)
		mLostEnergy += lostHP;
	//display vision noise
	if(mVisionNoise)
		displayNoise();
	//move away as a consequence of being hit
	mAuxiliaryV = direction * mConfig->getGameSpeed();
	mAuxVTime = 250;
	//inform the player
	mOGL->mOSD.addText(mTexts[HIT_TEXTS].getText());
	mSDL->playSound(mCollisionSounds[CSND_PLAYER_HIT], 1);
	//increase the heart beat, etc.
	excite();
}

void IGameCtrl::excite()
{
	mHeartBeat += 11.0f;
	if(mHeartBeat > 56.0f)
		mHeartBeat = 56.0f;
	mOGL->mOSD.setHeartSpeed(mHeartBeat);
	mExhaust = (mHeartBeat - 11.0f) / 45.0f;
}

bool IGameCtrl::equipmentCollision(Mesh3Dg *mesh, uint32_t dwPosOffset)
{
	//camera distaince from object
	Vector3Dg vec(mesh->mTranslation.R4C1, 0.0f, mesh->mTranslation.R4C3);
	vec = vec - mPrevCamPosition - mVelocity;
	//we don't have to calculate sqrtf - we can compare the dist^2 and ray^2
	float fDist = vec.lengthPow2();

	float fSpriteRay = mCamRay;
	//if equipment lays on a barrel or sth. then make the sprite ray longer
	if((mLevelMap[dwPosOffset + 4] & 16))
		fSpriteRay += 5.0f;
	//if distaince is greater than the sprite ray then there is no collision
	if(fDist <= fSpriteRay * fSpriteRay)
		return true;
	return false;
}

bool IGameCtrl::teleportCollision(Mesh3Dg *mesh, MapDataType wObjOffset, uint32_t &uiPastTime)
{
	//camera distaince from teleport
	Vector3Dg vec(mesh->mTranslation.R4C1, 0.0f, mesh->mTranslation.R4C3);
	vec = vec - mPrevCamPosition - mVelocity;
	//we don't have to calculate sqrtf - we can compare the dist^2 and ray^2
	float fDist = vec.lengthPow2();
	if(fDist <= (mCamRay * mCamRay)) {
		wObjOffset += 4;
		MapDataPtr pData = &mLevelMap[0x8000];
		uint32_t dwDataOffset = getDataOffset(wObjOffset, pData);
		//if it is finishing teleport
		if(dwDataOffset == 0)
			return false;
		dwDataOffset++;
		//if not working teleport (shouldn't happen, I think)
		if(pData[dwDataOffset] == 0x8888)
			return true;
		uint32_t dwZ = pData[dwDataOffset] / 512;
		float z = -(dwZ * 100.0f) - 50.0f,
			  x = ((pData[dwDataOffset] % 512) / 8) * 100.0f + 50.0f;
		teleportCamera(x, z);
		uiPastTime = teleportDissolve(x, z);
		closePrevDoor(wObjOffset, pData[dwDataOffset]);
	}
	return true;
}

uint32_t IGameCtrl::teleportDissolve(float x, float z)
{
	//redraw the most recent frame in position before teleportation
	mOGL->renderFrame();
	//copy the frame to the accumulation buffer
	mOGL->preDissolve(0.0f, 0.0f, 0.0f, DT_DISSOLVE);
	//teleport the camera and render the frame after teleportation
	mOGL->setCamMatrix(x, z, mSinCameraRotation, mCosCameraRotation);
	mOGL->renderFrame();
	uint32_t uiLastTime = mSDL->getTicks(), uiPastTime = 0;
	while(uiPastTime < 750) {
		uint32_t uiCurTime = mSDL->getTicks();
		uint32_t uiDiff = uiCurTime - uiLastTime;
		//frame rate limiter (max 100 fps during teleportation)
		//it helps controling the screen dissolve process
		if(uiDiff < 10)
			continue;
		uiLastTime = uiCurTime;
		//calculate the fade value
		float fFadeValue = 0.005 * (float)uiDiff;
		uiPastTime += uiDiff;
		mOGL->dissolveFrame(DT_DISSOLVE, fFadeValue);
		mOGL->mOSD.drawOSD(mSinCameraRotation, mCosCameraRotation, 0);
		mSDL->swapBuffers();
	}
	mSDL->clearEventQueue();
	return uiPastTime;
}

void IGameCtrl::teleportCamera(float x, float z)
{
	mCameraPosition.x = mPrevCamPosition.x = x;
	mCameraPosition.z = mPrevCamPosition.z = z;
	mPrevPosition[0] = mPosition[0];
	mPrevPosition[1] = mPosition[1];
	mPosition[0] = (uint8_t)(x / 100);
	mPosition[1] = (uint8_t)(-z / 100);
}

void IGameCtrl::damegeWeapon(Weapon *weapon)
{
	int minshots = mHighDificulty ? 100 : 200;
	float randomnumber = (rand() / (double)RAND_MAX) * (mHighDificulty ? 100.0f : 1000.0f);
	weapon->damage(minshots, randomnumber);
}

void IGameCtrl::useItem(UsableItem *pItem)
{
	switch(pItem->getType()) {
		case UIT_HAND:
			handCollision(pItem);
			break;
		case UIT_RED_CARD:
		case UIT_GREEN_CARD:
		case UIT_BLUE_CARD: {
			if(not ((Card *)pItem)->mCardsCount)
				mOGL->mOSD.addText(mTexts[NO_CARDS_TEXTS].getText());
			else if(cardCollision(pItem->getType())) {
				--((Card *)pItem)->mCardsCount;
				mSDL->playSound(pItem->mUseSoundNumber, 1);
			}
			break;
		}
		//regular weapons
		default: {
			//if the weapon is damaged
			if(pItem->getStatus() == WPNS_DAMAGED) {
				if(pItem->getType() == UIT_MASHINEGUN) {
					if(mMGFireDiff >= MACHINEGUN_FIRE_DELAY) {
						mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_WEAPON_DAMAGED));
						mSDL->playSound(((Weapon *)pItem)->mNoAmmoSoundNumber, 1);
						//delay next firing
						mMGFireDiff = 0;
					}
				} else {
					mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_WEAPON_DAMAGED));
					mSDL->playSound(((Weapon *)pItem)->mNoAmmoSoundNumber, 1);
				}
				break;
			}
			//if no ammo
			//(for mashinegun we handle the lack of ammo in a bit different way)
			if(not ((Weapon *)pItem)->getAmmo()->getAmount() and pItem->getType() != UIT_MASHINEGUN) {
				mOGL->mOSD.addText(mTexts[NO_AMMO_TEXTS].getText());
				mSDL->playSound(((Weapon *)pItem)->mNoAmmoSoundNumber, 1);
				break;
			}
			switch(pItem->getType()) {
				case UIT_HANDGUN:
				case UIT_SHOTGUN: {
					Weapon *weapon = (Weapon *)pItem;
					weapon->getAmmo()->use();
					damegeWeapon(weapon);
					mSDL->playSound(weapon->mUseSoundNumber, 1);
					Vector3Dg pos = mCameraPosition;
					pos.y = 50.0f;
					for(uint16_t i = 0; i < weapon->getBullets(); i++)
						ammoCollision(weapon, pos, mCameraDirection);
					break;
				}
				case UIT_MASHINEGUN: {
					if(mMGFireDiff >= MACHINEGUN_FIRE_DELAY) {
						//delay next firing
						mMGFireDiff = 0;
						Weapon *weapon = (Weapon *)pItem;
						if(not weapon->getAmmo()->getAmount()) {
							mOGL->mOSD.addText(mTexts[NO_AMMO_TEXTS].getText());
							mSDL->playSound(weapon->mNoAmmoSoundNumber, 1);
							break;
						}
						weapon->getAmmo()->use();
						damegeWeapon(weapon);
						mSDL->playSound(weapon->mUseSoundNumber, 1);
						Vector3Dg pos = mCameraPosition;
						pos.y = 50.0f;
						for(uint16_t i = 0; i < weapon->getBullets(); i++)
							ammoCollision(weapon, pos, mCameraDirection);
					}
					break;
				}
				case UIT_FLAMER:
				case UIT_BLASTER: {
					Weapon *weapon = (Weapon *)pItem;
					weapon->getAmmo()->use();
					damegeWeapon(weapon);
					mSDL->playSound(pItem->mUseSoundNumber, 1);
					Vector3Dg pos = mCameraPosition + (mCameraDirection * 25.0f);
					fireAmmo(weapon, pos, mCameraDirection);
					break;
				}
				case UIT_ROCKETLAUNCHER: {
					Weapon *weapon = (Weapon *)pItem;
					if(mSDL->getTicks() - mRocketLoadTime < 1000)
						mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_LOADING_LAUNCHER));
					else {
						weapon->getAmmo()->use();
						damegeWeapon(weapon);
						Vector3Dg pos = mCameraPosition + (mCameraDirection * 25.0f);
						fireAmmo(weapon, pos, mCameraDirection);
						mSDL->playSound(pItem->mUseSoundNumber, 1);
						//delay next firing
						mRocketLoadTime = mSDL->getTicks();
					}
					break;
				}
				default:
					break;
			}
			break;
		}
	}
}

uint32_t IGameCtrl::getDataOffset(MapDataType wObjOffset, MapDataPtr pData)
{
	uint32_t dwDataSize = (mObjects->getMapDataSize() / 2) - 0x8000;PRE_LOOP();
	for(uint32_t i = 1; i < dwDataSize; i++) {
		if(pData[i] == wObjOffset and pData[i - 1] == 0xFFFF)
		{AFTER_LOOP();
			return i;}
	}AFTER_LOOP();
	return 0;
}

int32_t IGameCtrl::calculateVolume(Vector3Dg position, VolumeCalculationMethod method)
{
	Vector3Dg vec = position - mCameraPosition;
	float distaince = vec.length();
	float max = 0.0f;
	switch(method) {
		case VOLDIST_TYPE1:
			max = 1000.0f;
			break;
		case VOLDIST_TYPE2:
			max = 800.0f;
			break;
		case VOLDIST_TYPE3:
			max = 700.0f;
			break;
	}

	if(distaince > max)
		return 0;
	distaince *= 128.0f / max;
	return 128 - (int32_t)distaince;
}

void IGameCtrl::displayNoise()
{
	uint32_t dwNoiseLines[10];
	for(int32_t i = 0; i < 10; ++i)
		dwNoiseLines[i] = rand();
	mOGL->mOSD.addNoise(dwNoiseLines);
}

uint16_t IGameCtrl::getNumberOfKilledEnemies()
{
	uint16_t counter = 0;
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++)
		if(not ((GOEnemy *)*i)->isAlive())
			counter++;

	return counter;
}

void IGameCtrl::leaveAmmo(GOEnemy *enemy)
{
	//leave the ammo randomly (the probability is 0.25)
	if(not randomTrue(0.25f))
		return;

	//check if there is a place for ammo
	Vector3Dg pos = enemy->getMesh()->geomCenter();
	uint32_t x = ((int)pos.x/100);
	uint32_t z = -((int)pos.z/100);
	uint32_t posOffset = x * 8 + z * 512;
	if(mLevelMap[posOffset + 4] & 32)
		return;

	//is there already an equipment mesh?
	Mesh3DgUID uid = mLevelMap[posOffset + 6];
	Mesh3Dg *mesh;
	if(uid == 0) {
		uid = mObjects->generateMeshUID();
		mesh = mObjects->createMesh(uid);
		//no memory? no big deal - just don't create the equipment
		if(mesh == 0)
			return;
		mesh->setProperty(MRP_TRANSPARENT, true);
		//game object
		if(mObjects->createGameObject(GOTYPE_SPRITE, mesh) == 0) {
			mObjects->destroyMesh(uid);
			return;
		}
		//set the position
		pos.x = ((int)pos.x/100) * 100.0f + 50.0f;
		pos.z = ((int)pos.z/100) * 100.0f - 50.0f;
		mesh->setPosition(pos.x, 25.0f, pos.z);
		mesh->setBase(mEquipmentBase);
		mOGL->addMeshToRenderList(mesh);
		mLevelMap[posOffset + 6] = uid;
	} else {
		mesh = mObjects->mMeshes[uid];
		mesh->setProperty(MRP_RENDERABLE, true);
	}

	//the equipment type
	mesh->mMaterialNumber = 61 + enemy->getWeaponType();
	//info about equipment on the map
	mLevelMap[posOffset + 4] |= 32;
}

void IGameCtrl::spreadBlood(GOEnemy *enemy, Vector3Dg direction, float shooterDistaince)
{
	//calculate the position offset in map
	Vector3Dg pos = enemy->getMesh()->geomCenter();
	uint32_t x = ((int)pos.x/100);
	uint32_t z = -((int)pos.z/100);
	uint32_t posOffset = x * 8 + z * 512;

	//check which walls are present on this field
	MapDataType walls = mLevelMap[posOffset + 4];
	for(uint32_t i = 0; i < 4; ++i) {
		if(walls & (1 << i)) {
			//calculate the dot product of the direction vector and the wall's normal vector:
			//the normal vector of the wall
			Vector3Dg normal(0.0f, 0.0f, 0.0f);
			switch(i) {
				case 0: normal.z = 1.0f; break;
				case 1: normal.x = -1.0f; break;
				case 2: normal.z = -1.0f; break;
				case 3: normal.x = 1.0f; break;
			}
			//the dot product - dependant on it's value, we can determine the wall's position
			//relative to the shooter
			float dotP = direction.dotProduct(normal);
			//if the wall is behind the enemy - spread the blood on this wall (choose one of
			//two blood textures that are suitable for this situation)
			if(dotP < -SQR2BY2) {
				//set the blood material
				GameObject *obj = mObjects->getGameObject(mLevelMap[posOffset + i]);
				if(obj->getType() != GOTYPE_WALL)
					continue;
				((GOWall *)obj)->setBlood(randomTrue(0.5f) ? 0 : 1);
				continue;
			}

			//the wall is in front of the enemy
			if(dotP > SQR2BY2)
				continue;

			//if the wall is at the side (only one blood texture for spreading on side walls)
			//spread the blood with some probability, dependant on the shooter's distainde from
			//the enemy
			if(not randomTrue((1000.0f - shooterDistaince) / 1000.0f))
				continue;
			GameObject *obj = mObjects->getGameObject(mLevelMap[posOffset + i]);
			if(obj->getType() != GOTYPE_WALL)
				continue;
			((GOWall *)obj)->setBlood(2);
		}
	}
}

bool IGameCtrl::randomTrue(float probability)
{
	float randomnumber = (rand() / (double)RAND_MAX);
	if(randomnumber < probability or probability >= 1.0)
		return true;
	else return false;
}

void IGameCtrl::unregisterFireSounds()
{
	for(EnemiesList::iterator i = mEnemies.begin(); i != mEnemies.end(); i++)
		((GOEnemy *)*i)->unregisterSound();
}

void IGameCtrl::randomRoar(GOEnemy *enemy)
{
	if(not randomTrue(0.008))
		return;
	Vector3Dg pos = enemy->getMesh()->geomCenter();
	int32_t volume = calculateVolume(pos, VOLDIST_TYPE3);
	if(volume == 0)
		return;

	int num = (int)((rand() / (double)RAND_MAX) * 3.0) % 3;
	mSDL->playSound(mEnemySounds[ESND_ROAR1 + num], 4, volume);
}

void IGameCtrl::hitRoar(GOEnemy *enemy)
{
	if(not randomTrue(0.5))
		return;

	Vector3Dg pos = enemy->getMesh()->geomCenter();
	mSDL->playSound(mEnemySounds[ESND_ROAR2], 4, calculateVolume(pos, VOLDIST_TYPE3));
}

void IGameCtrl::checkEnemyCollisions(GOEnemy *enemy)
{
	Vector3Dg pos = enemy->getMesh()->geomCenter();
	int8_t enemyPosX = (int8_t)((uint32_t)pos.x/100);
	int8_t enemyPosZ = (int8_t)((uint32_t)-pos.z/100);
	uint32_t posOffset = enemyPosX * 8 + enemyPosZ * 512;

	//collisions with walls, door and edges
	bool edges[] = {false, false, false, false, false};
	//get info about existing walls in the square
	uint16_t walls = mLevelMap[posOffset + 4];
	for(uint32_t i = 0; i < 4; ++i) {
		if(walls & (1 << i)) {
			Mesh3DgUID meshUID = mLevelMap[posOffset + i];
			edges[i] = edges[i + 1] = true;
			GameObject *obj = mObjects->getGameObject(meshUID);
			if(obj->getPlayerCollisionType() != PCTYPE_BUMP)
				continue;
			switch(obj->getType()) {
				case GOTYPE_WALL:
					enemyWallCollision(enemy, obj->getMesh(), i, mEnemyRay);
					break;
				case GOTYPE_DOOR: {
					//only if not open
					if(isDoorOpen((GODoor *)obj))
						enemyWallCollision(enemy, obj->getMesh(), i, mEnemyRay);
					break;
				}
				default:
					break;
			}
		} else {
			//check if there is a wall on the neighbouring field
			uint16_t newPosOffset;
			switch(i) {
				case 0:
					newPosOffset = getRelativePosOffset(enemyPosX, enemyPosZ, 0, 1);
					break;
				case 1:
					newPosOffset = getRelativePosOffset(enemyPosX, enemyPosZ, 1, 0);
					break;
				case 2:
					newPosOffset = getRelativePosOffset(enemyPosX, enemyPosZ, 0, -1);
					break;
				case 3:
					newPosOffset = getRelativePosOffset(enemyPosX, enemyPosZ, -1, 0);
					break;
			}
			uint16_t newWalls = mLevelMap[newPosOffset + 4];
			//prevent the enemy from going through the wall even from the back side
			uint32_t j = (i + 2) % 4;
			if(newWalls & (1 << j)) {
				Mesh3DgUID meshUID = mLevelMap[newPosOffset + j];
				GameObject *obj = mObjects->getGameObject(meshUID);
				if(obj->getPlayerCollisionType() != PCTYPE_BUMP)
					continue;
				enemyBackWallCollision(enemy, obj->getMesh(), j, mEnemyRay);
			}
		}
	}
	edges[0] = edges[0] or edges[4];
	searchForEdges(edges, enemyPosX, enemyPosZ);
	//check edges collisions
	enemyEdgesCollision(edges, enemy, enemyPosX, enemyPosZ, mEnemyRay);

	//collision with player
	if(enemyObjectCollision(mCameraPosition, enemy, mCamRay, mEnemyRay)) {
		//if the enemy is under fire then the player must get burnt
		if(enemy->isBurning() and mPlayerBurnDiff >= PLAYER_BURN_DELAY) {
			mPlayerBurnDiff = 0;
			mOGL->mOSD.addText(mLocInterface->getText(LOC_GAME_HOT));
			mSDL->playSound(mCollisionSounds[CSND_TOUCH_FLAMES], 1);
			excite();
			if(mHighDificulty and not mImmortal)
				mLostEnergy++;
		}
	}

	//collision with other enemies
	enemyCollision(enemy);
	//don't check other collisions if enemy is outside the level map
	if(pos.x > 6400.0f or pos.x < 0.0f or pos.z < -6400.0f or pos.z > 0.0f)
		return;

	//collision with sprites
	if(mLevelMap[posOffset + 4] & 16) {
		Mesh3DgUID meshUID = mLevelMap[posOffset + 5];
		Mesh3Dg *mesh = mObjects->mMeshes[meshUID];
		if(((GameObject *)mesh->getChild())->getPlayerCollisionType() == PCTYPE_STOP) {
			pos = mesh->geomCenter();
			enemyObjectCollision(pos, enemy, 0.0f, mEnemyRay);
		}
	}
}

bool IGameCtrl::canShoot(float range, Vector3Dg position, Vector3Dg direction)
{
	//calculate the end point of the ammo's way
	Vector3Dg end = position + (direction * range);
	//random shift in direction
	float shift = 25.0f - (rand() / (double)RAND_MAX) * 50.0f;
	end.x += direction.z * shift;
	shift = 25.0f - (rand() / (double)RAND_MAX) * 50.0f;
	end.z += direction.x * shift;
	//calculate the direction once again 'cause we gonna need it
	direction = end - position;
	direction.normalize();
	//calculate the boundaries of potentially affected area
	int8_t beginx = (int8_t)(position.x / 100), beginz = (int8_t)(-position.z / 100);
	int8_t endx = (int8_t)(end.x / 100), endz = (int8_t)(-end.z / 100);
	//make sure that begining values are lesser than ending ones
	if(endx < 0)
		endx = 0;
	if(endz < 0)
		endz = 0;
	if(beginx > endx) {
		int8_t tmp = beginx;
		beginx = endx;
		endx = tmp;
	}
	if(beginz > endz) {
		int8_t tmp = beginz;
		beginz = endz;
		endz = tmp;
	}
	//make sure we are not out of bounds
	if(beginx < 0)
		beginx = 0;
	if(beginz < 0)
		beginz = 0;
	if(endx > 63)
		endx = 63;
	if(endz > 63)
		endz = 63;

	//which walls to check
	uint32_t count = 0, check[2];
	//north
	if(direction.z < 0.0f) {
		check[count] = 0;
		count++;
	}
	//south
	else if(direction.z > 0.0f) {
		check[count] = 2;
		count++;
	}
	//east
	if(direction.x > 0.0f) {
		check[count] = 1;
		count++;
	}
	//west
	else if(direction.x < 0.0f) {
		check[count] = 3;
		count++;
	}

	//distaince of the closest object
	float distaince = range;
	GameObject *closestobject = 0;

	//check all fields for collisions
	for(int8_t c = beginx; c <= endx; c++) {
		for(int8_t r = beginz; r <= endz; r++) {
			//position's offset in the map buffer
			uint32_t offset = c * 8 + r * 512;

			//get info about existing walls in the square
			uint16_t walls = mLevelMap[offset + 4];
			for(uint32_t i = 0; i < count; ++i) {
				if(not (walls & (1 << check[i])))
					continue;
				Mesh3DgUID meshnum = mLevelMap[offset + check[i]];
				GameObject *obj = mObjects->getGameObject(meshnum);
				if(obj->getAmmoCollisionType() == ACTYPE_NONE)
					continue;
				float dist = 0.0f;
				switch(obj->getType()) {
					case GOTYPE_WALL: {
						if(((GOWall *)obj)->getTransparentForEnemy())
							break;
						dist = wallCollisionAmmo(position, end, c, r, check[i], obj);
						break;
					}
					case GOTYPE_DOOR: {
						//only if not open
						if(isDoorOpen((GODoor *)obj))
							dist = wallCollisionAmmo(position, end, c, r, check[i], obj);
						break;
					}
					default:
						break;
				}
				if(dist != 0.0f and dist < distaince) {
					distaince = dist;
					closestobject = obj;
				}
			}

			//sprites
			Mesh3DgUID meshnum = mLevelMap[offset + 5];
			if(mLevelMap[offset + 4] & 16) {
				GameObject *obj = mObjects->getGameObject(meshnum);
				if(obj->getAmmoCollisionType() == ACTYPE_UNBREAKABLE) {
					float dist = spriteCollisionAmmo(position, end, c, r, range);
					if(dist != 0.0f and dist < distaince) {
						distaince = dist;
						closestobject = obj;
					}
				}
			}
		}
	}

	if(closestobject == 0)
		return true;
	else return false;
}
